"""
Nexa SDK - Python wrapper for nexa-sdk C library.

This module provides low-level ctypes bindings to the nexa-sdk C library.
All Python implementations are in nexaai and internal packages.
"""

# Export library loader
from ._lib import _lib

# Export ASR structures and types
from .asr import (
    ml_ASR,
    ml_asr_transcription_callback,
    ml_ASRConfig,
    ml_AsrCreateInput,
    ml_AsrListSupportedLanguagesInput,
    ml_AsrListSupportedLanguagesOutput,
    ml_ASRResult,
    ml_AsrStreamBeginInput,
    ml_AsrStreamBeginOutput,
    ml_ASRStreamConfig,
    ml_AsrStreamPushAudioInput,
    ml_AsrStreamStopInput,
    ml_AsrTranscribeInput,
    ml_AsrTranscribeOutput,
)

# Export CV structures and types
from .cv import (
    ml_BoundingBox,
    ml_CV,
    ml_CVCapabilities,
    ml_CVCreateInput,
    ml_CVInferInput,
    ml_CVInferOutput,
    ml_CVModelConfig,
    ml_CVResult,
)

# Export Diarize structures and types
from .diarize import (
    ml_Diarize,
    ml_DiarizeConfig,
    ml_DiarizeCreateInput,
    ml_DiarizeInferInput,
    ml_DiarizeInferOutput,
    ml_DiarizeSpeechSegment,
)

# Export Embedding structures and types
from .embedding import (
    ml_Embedder,
    ml_EmbedderCreateInput,
    ml_EmbedderDimOutput,
    ml_EmbedderEmbedInput,
    ml_EmbedderEmbedOutput,
    ml_EmbeddingConfig,
)

# Export error codes and definitions
# Export error classes (from error.py)
from .error import (
    _ERROR_MAP,
    ML_ERROR_COMMON_FILE_NOT_FOUND,
    ML_ERROR_COMMON_INVALID_INPUT,
    ML_ERROR_COMMON_LICENSE_EXPIRED,
    ML_ERROR_COMMON_LICENSE_INVALID,
    ML_ERROR_COMMON_MEMORY_ALLOCATION,
    ML_ERROR_COMMON_MODEL_INVALID,
    ML_ERROR_COMMON_MODEL_LOAD,
    ML_ERROR_COMMON_NOT_INITIALIZED,
    ML_ERROR_COMMON_NOT_SUPPORTED,
    ML_ERROR_COMMON_UNKNOWN,
    ML_ERROR_LLM_GENERATION_FAILED,
    ML_ERROR_LLM_GENERATION_PROMPT_TOO_LONG,
    ML_ERROR_LLM_TOKENIZATION_CONTEXT_LENGTH,
    ML_ERROR_LLM_TOKENIZATION_FAILED,
    ML_LOG_LEVEL_DEBUG,
    ML_LOG_LEVEL_ERROR,
    ML_LOG_LEVEL_INFO,
    ML_LOG_LEVEL_TRACE,
    ML_LOG_LEVEL_WARN,
    ML_SUCCESS,
    NexaCommonError,
    NexaError,
    NexaFileNotFoundError,
    NexaInvalidInputError,
    NexaLicenseExpiredError,
    NexaLicenseInvalidError,
    NexaLLMError,
    NexaLLMGenerationError,
    NexaLLMGenerationPromptTooLongError,
    NexaLLMTokenizationContextLengthError,
    NexaLLMTokenizationError,
    NexaMemoryAllocationError,
    NexaModelInvalidError,
    NexaModelLoadError,
    NexaNotInitializedError,
    NexaNotSupportedError,
    NexaUnknownError,
    check_error,
    ml_get_error_message,
    ml_log_callback,
)

# Export ImageGen structures and types
from .image_gen import (
    ml_ImageGen,
    ml_ImageGenCreateInput,
    ml_ImageGenerationConfig,
    ml_ImageGenImg2ImgInput,
    ml_ImageGenOutput,
    ml_ImageGenTxt2ImgInput,
    ml_ImageSamplerConfig,
    ml_SchedulerConfig,
)

# Export LLM structures and types
from .llm import (
    ml_GenerationConfig,
    ml_KvCacheLoadInput,
    ml_KvCacheLoadOutput,
    ml_KvCacheSaveInput,
    ml_KvCacheSaveOutput,
    ml_LLM,
    ml_LlmApplyChatTemplateInput,
    ml_LlmApplyChatTemplateOutput,
    ml_LlmChatMessage,
    ml_LlmCreateInput,
    ml_LlmGenerateInput,
    ml_LlmGenerateOutput,
    ml_ModelConfig,
    ml_ProfileData,
    ml_SamplerConfig,
    ml_token_callback,
)

# Export ML core structures and types
from .ml import (
    ml_create_plugin_func,
    ml_GetDeviceListInput,
    ml_GetDeviceListOutput,
    ml_GetPluginListOutput,
    ml_plugin_id_func,
)

# Export Rerank structures and types
from .rerank import (
    ml_RerankConfig,
    ml_Reranker,
    ml_RerankerCreateInput,
    ml_RerankerRerankInput,
    ml_RerankerRerankOutput,
)

# Export TTS structures and types
from .tts import (
    ml_TTS,
    ml_TTSConfig,
    ml_TtsCreateInput,
    ml_TtsListAvailableVoicesInput,
    ml_TtsListAvailableVoicesOutput,
    ml_TTSResult,
    ml_TTSSamplerConfig,
    ml_TtsSynthesizeInput,
    ml_TtsSynthesizeOutput,
)

# Export types
from .types import (
    GenerationConfig,
    KvCacheLoadInput,
    KvCacheSaveInput,
    LlmChatMessage,
    ModelConfig,
    ProfileData,
    SamplerConfig,
)

# Export VLM structures and types
from .vlm import (
    ml_VLM,
    ml_VlmApplyChatTemplateInput,
    ml_VlmApplyChatTemplateOutput,
    ml_VlmChatMessage,
    ml_VlmContent,
    ml_VlmCreateInput,
    ml_VlmGenerateInput,
    ml_VlmGenerateOutput,
)

__version__ = '0.1.0'

__all__ = [
    # Library
    '_lib',
    # Error codes
    'ML_SUCCESS',
    'ML_ERROR_COMMON_UNKNOWN',
    'ML_ERROR_COMMON_INVALID_INPUT',
    'ML_ERROR_COMMON_MEMORY_ALLOCATION',
    'ML_ERROR_COMMON_FILE_NOT_FOUND',
    'ML_ERROR_COMMON_NOT_INITIALIZED',
    'ML_ERROR_COMMON_NOT_SUPPORTED',
    'ML_ERROR_COMMON_MODEL_LOAD',
    'ML_ERROR_COMMON_MODEL_INVALID',
    'ML_ERROR_COMMON_LICENSE_INVALID',
    'ML_ERROR_COMMON_LICENSE_EXPIRED',
    'ML_ERROR_LLM_TOKENIZATION_FAILED',
    'ML_ERROR_LLM_TOKENIZATION_CONTEXT_LENGTH',
    'ML_ERROR_LLM_GENERATION_FAILED',
    'ML_ERROR_LLM_GENERATION_PROMPT_TOO_LONG',
    'ML_LOG_LEVEL_TRACE',
    'ML_LOG_LEVEL_DEBUG',
    'ML_LOG_LEVEL_INFO',
    'ML_LOG_LEVEL_WARN',
    'ML_LOG_LEVEL_ERROR',
    'ml_log_callback',
    # ML core types
    'ml_plugin_id_func',
    'ml_create_plugin_func',
    'ml_GetPluginListOutput',
    'ml_GetDeviceListInput',
    'ml_GetDeviceListOutput',
    # LLM types
    'ml_SamplerConfig',
    'ml_GenerationConfig',
    'ml_ModelConfig',
    'ml_LLM',
    'ml_LlmCreateInput',
    'ml_LlmChatMessage',
    'ml_LlmApplyChatTemplateInput',
    'ml_LlmApplyChatTemplateOutput',
    'ml_ProfileData',
    'ml_LlmGenerateInput',
    'ml_LlmGenerateOutput',
    'ml_KvCacheSaveInput',
    'ml_KvCacheSaveOutput',
    'ml_KvCacheLoadInput',
    'ml_KvCacheLoadOutput',
    'ml_token_callback',
    # VLM types
    'ml_VLM',
    'ml_VlmContent',
    'ml_VlmChatMessage',
    'ml_VlmCreateInput',
    'ml_VlmApplyChatTemplateInput',
    'ml_VlmApplyChatTemplateOutput',
    'ml_VlmGenerateInput',
    'ml_VlmGenerateOutput',
    # ASR types
    'ml_ASR',
    'ml_ASRConfig',
    'ml_ASRResult',
    'ml_ASRStreamConfig',
    'ml_AsrCreateInput',
    'ml_AsrTranscribeInput',
    'ml_AsrTranscribeOutput',
    'ml_AsrListSupportedLanguagesInput',
    'ml_AsrListSupportedLanguagesOutput',
    'ml_AsrStreamBeginInput',
    'ml_AsrStreamBeginOutput',
    'ml_AsrStreamPushAudioInput',
    'ml_AsrStreamStopInput',
    'ml_asr_transcription_callback',
    # Embedding types
    'ml_Embedder',
    'ml_EmbeddingConfig',
    'ml_EmbedderCreateInput',
    'ml_EmbedderEmbedInput',
    'ml_EmbedderEmbedOutput',
    'ml_EmbedderDimOutput',
    # TTS types
    'ml_TTS',
    'ml_TTSConfig',
    'ml_TTSSamplerConfig',
    'ml_TTSResult',
    'ml_TtsCreateInput',
    'ml_TtsSynthesizeInput',
    'ml_TtsSynthesizeOutput',
    'ml_TtsListAvailableVoicesInput',
    'ml_TtsListAvailableVoicesOutput',
    # Diarize types
    'ml_Diarize',
    'ml_DiarizeConfig',
    'ml_DiarizeSpeechSegment',
    'ml_DiarizeCreateInput',
    'ml_DiarizeInferInput',
    'ml_DiarizeInferOutput',
    # Rerank types
    'ml_Reranker',
    'ml_RerankConfig',
    'ml_RerankerCreateInput',
    'ml_RerankerRerankInput',
    'ml_RerankerRerankOutput',
    # CV types
    'ml_CV',
    'ml_BoundingBox',
    'ml_CVResult',
    'ml_CVCapabilities',
    'ml_CVModelConfig',
    'ml_CVCreateInput',
    'ml_CVInferInput',
    'ml_CVInferOutput',
    # ImageGen types
    'ml_ImageGen',
    'ml_ImageSamplerConfig',
    'ml_SchedulerConfig',
    'ml_ImageGenerationConfig',
    'ml_ImageGenCreateInput',
    'ml_ImageGenTxt2ImgInput',
    'ml_ImageGenImg2ImgInput',
    'ml_ImageGenOutput',
    # Error classes
    'NexaError',
    'NexaCommonError',
    'NexaUnknownError',
    'NexaInvalidInputError',
    'NexaMemoryAllocationError',
    'NexaFileNotFoundError',
    'NexaNotInitializedError',
    'NexaNotSupportedError',
    'NexaModelLoadError',
    'NexaModelInvalidError',
    'NexaLicenseInvalidError',
    'NexaLicenseExpiredError',
    'NexaLLMError',
    'NexaLLMTokenizationError',
    'NexaLLMTokenizationContextLengthError',
    'NexaLLMGenerationError',
    'NexaLLMGenerationPromptTooLongError',
    # Error handling functions
    'check_error',
    'ml_get_error_message',
    # Types
    'SamplerConfig',
    'GenerationConfig',
    'ModelConfig',
    'LlmChatMessage',
    'ProfileData',
    'KvCacheSaveInput',
    'KvCacheLoadInput',
]



"""
Image generation functions and structures for nexa-sdk.
"""

import ctypes

from ._lib import _lib
from .error import ml_log_callback
from .llm import ml_ModelConfig

# ============================================================================
# C Structure Definitions
# ============================================================================


class ml_ImageSamplerConfig(ctypes.Structure):
    """Image generation sampling parameters."""

    _fields_ = [
        ('method', ctypes.c_char_p),
        ('steps', ctypes.c_int32),
        ('guidance_scale', ctypes.c_float),
        ('eta', ctypes.c_float),
        ('seed', ctypes.c_int32),
    ]


class ml_SchedulerConfig(ctypes.Structure):
    """Diffusion scheduler configuration."""

    _fields_ = [
        ('type', ctypes.c_char_p),
        ('num_train_timesteps', ctypes.c_int32),
        ('steps_offset', ctypes.c_int32),
        ('beta_start', ctypes.c_float),
        ('beta_end', ctypes.c_float),
        ('beta_schedule', ctypes.c_char_p),
        ('prediction_type', ctypes.c_char_p),
        ('timestep_type', ctypes.c_char_p),
        ('timestep_spacing', ctypes.c_char_p),
        ('interpolation_type', ctypes.c_char_p),
        ('config_path', ctypes.c_char_p),
    ]


class ml_ImageGenerationConfig(ctypes.Structure):
    """Image generation configuration."""

    _fields_ = [
        ('prompts', ctypes.POINTER(ctypes.c_char_p)),
        ('prompt_count', ctypes.c_int32),
        ('negative_prompts', ctypes.POINTER(ctypes.c_char_p)),
        ('negative_prompt_count', ctypes.c_int32),
        ('height', ctypes.c_int32),
        ('width', ctypes.c_int32),
        ('sampler_config', ml_ImageSamplerConfig),
        ('scheduler_config', ml_SchedulerConfig),
        ('strength', ctypes.c_float),
    ]


class ml_ImageGen(ctypes.Structure):
    """Opaque image generator handle."""

    pass


class ml_ImageGenCreateInput(ctypes.Structure):
    """Input structure for creating an image generator."""

    _fields_ = [
        ('model_name', ctypes.c_char_p),
        ('model_path', ctypes.c_char_p),
        ('config', ml_ModelConfig),
        ('scheduler_config_path', ctypes.c_char_p),
        ('plugin_id', ctypes.c_char_p),
        ('device_id', ctypes.c_char_p),
    ]


class ml_ImageGenTxt2ImgInput(ctypes.Structure):
    """Input structure for text-to-image generation."""

    _fields_ = [
        ('prompt_utf8', ctypes.c_char_p),
        ('config', ctypes.POINTER(ml_ImageGenerationConfig)),
        ('output_path', ctypes.c_char_p),
    ]


class ml_ImageGenImg2ImgInput(ctypes.Structure):
    """Input structure for image-to-image generation."""

    _fields_ = [
        ('init_image_path', ctypes.c_char_p),
        ('prompt_utf8', ctypes.c_char_p),
        ('config', ctypes.POINTER(ml_ImageGenerationConfig)),
        ('output_path', ctypes.c_char_p),
    ]


class ml_ImageGenOutput(ctypes.Structure):
    """Output structure for image generation."""

    _fields_ = [
        ('output_image_path', ctypes.c_char_p),
    ]


# ============================================================================
# Function Signatures
# ============================================================================

_lib.ml_imagegen_create.argtypes = [
    ctypes.POINTER(ml_ImageGenCreateInput),
    ctypes.POINTER(ctypes.POINTER(ml_ImageGen)),
]
_lib.ml_imagegen_create.restype = ctypes.c_int32

_lib.ml_imagegen_destroy.argtypes = [ctypes.POINTER(ml_ImageGen)]
_lib.ml_imagegen_destroy.restype = ctypes.c_int32

_lib.ml_imagegen_txt2img.argtypes = [
    ctypes.POINTER(ml_ImageGen),
    ctypes.POINTER(ml_ImageGenTxt2ImgInput),
    ctypes.POINTER(ml_ImageGenOutput),
]
_lib.ml_imagegen_txt2img.restype = ctypes.c_int32

_lib.ml_imagegen_img2img.argtypes = [
    ctypes.POINTER(ml_ImageGen),
    ctypes.POINTER(ml_ImageGenImg2ImgInput),
    ctypes.POINTER(ml_ImageGenOutput),
]
_lib.ml_imagegen_img2img.restype = ctypes.c_int32



"""
ASR functions and structures for nexa-sdk.
"""

import ctypes

from ._lib import _lib
from .error import ml_log_callback
from .llm import ml_ModelConfig, ml_ProfileData

# ============================================================================
# C Structure Definitions
# ============================================================================

ml_asr_transcription_callback = ctypes.CFUNCTYPE(None, ctypes.c_char_p, ctypes.c_void_p)


class ml_ASRConfig(ctypes.Structure):
    """ASR processing configuration."""

    _fields_ = [
        ('timestamps', ctypes.c_char_p),
        ('beam_size', ctypes.c_int32),
        ('stream', ctypes.c_bool),
    ]


class ml_ASRResult(ctypes.Structure):
    """ASR transcription result."""

    _fields_ = [
        ('transcript', ctypes.c_char_p),
        ('confidence_scores', ctypes.POINTER(ctypes.c_float)),
        ('confidence_count', ctypes.c_int32),
        ('timestamps', ctypes.POINTER(ctypes.c_float)),
        ('timestamp_count', ctypes.c_int32),
    ]


class ml_ASR(ctypes.Structure):
    """Opaque ASR handle."""

    pass


class ml_AsrCreateInput(ctypes.Structure):
    """Input structure for creating an ASR instance."""

    _fields_ = [
        ('model_name', ctypes.c_char_p),
        ('model_path', ctypes.c_char_p),
        ('tokenizer_path', ctypes.c_char_p),
        ('config', ml_ModelConfig),
        ('language', ctypes.c_char_p),
        ('plugin_id', ctypes.c_char_p),
        ('device_id', ctypes.c_char_p),
        ('license_id', ctypes.c_char_p),
        ('license_key', ctypes.c_char_p),
    ]


class ml_AsrTranscribeInput(ctypes.Structure):
    """Input structure for ASR transcription."""

    _fields_ = [
        ('audio_path', ctypes.c_char_p),
        ('language', ctypes.c_char_p),
        ('config', ctypes.POINTER(ml_ASRConfig)),
    ]


class ml_AsrTranscribeOutput(ctypes.Structure):
    """Output structure for ASR transcription."""

    _fields_ = [
        ('result', ml_ASRResult),
        ('profile_data', ml_ProfileData),
    ]


class ml_AsrListSupportedLanguagesInput(ctypes.Structure):
    """Input structure for getting supported languages."""

    _fields_ = [
        ('reserved', ctypes.c_void_p),
    ]


class ml_AsrListSupportedLanguagesOutput(ctypes.Structure):
    """Output structure for getting supported languages."""

    _fields_ = [
        ('language_codes', ctypes.POINTER(ctypes.c_char_p)),
        ('language_count', ctypes.c_int32),
    ]


class ml_ASRStreamConfig(ctypes.Structure):
    """ASR streaming configuration."""

    _fields_ = [
        ('chunk_duration', ctypes.c_float),
        ('overlap_duration', ctypes.c_float),
        ('sample_rate', ctypes.c_int32),
        ('max_queue_size', ctypes.c_int32),
        ('buffer_size', ctypes.c_int32),
        ('timestamps', ctypes.c_char_p),
        ('beam_size', ctypes.c_int32),
    ]


class ml_AsrStreamBeginInput(ctypes.Structure):
    """Input structure for beginning ASR streaming."""

    _fields_ = [
        ('stream_config', ctypes.POINTER(ml_ASRStreamConfig)),
        ('language', ctypes.c_char_p),
        ('on_transcription', ml_asr_transcription_callback),
        ('user_data', ctypes.c_void_p),
    ]


class ml_AsrStreamBeginOutput(ctypes.Structure):
    """Output structure for streaming begin."""

    _fields_ = [
        ('reserved', ctypes.c_void_p),
    ]


class ml_AsrStreamPushAudioInput(ctypes.Structure):
    """Input structure for processing audio data."""

    _fields_ = [
        ('audio_data', ctypes.POINTER(ctypes.c_float)),
        ('length', ctypes.c_int32),
    ]


class ml_AsrStreamStopInput(ctypes.Structure):
    """Input structure for stopping streaming."""

    _fields_ = [
        ('graceful', ctypes.c_bool),
    ]


# ============================================================================
# Function Signatures
# ============================================================================

_lib.ml_asr_create.argtypes = [ctypes.POINTER(ml_AsrCreateInput), ctypes.POINTER(ctypes.POINTER(ml_ASR))]
_lib.ml_asr_create.restype = ctypes.c_int32

_lib.ml_asr_destroy.argtypes = [ctypes.POINTER(ml_ASR)]
_lib.ml_asr_destroy.restype = ctypes.c_int32

_lib.ml_asr_transcribe.argtypes = [
    ctypes.POINTER(ml_ASR),
    ctypes.POINTER(ml_AsrTranscribeInput),
    ctypes.POINTER(ml_AsrTranscribeOutput),
]
_lib.ml_asr_transcribe.restype = ctypes.c_int32

_lib.ml_asr_list_supported_languages.argtypes = [
    ctypes.POINTER(ml_ASR),
    ctypes.POINTER(ml_AsrListSupportedLanguagesInput),
    ctypes.POINTER(ml_AsrListSupportedLanguagesOutput),
]
_lib.ml_asr_list_supported_languages.restype = ctypes.c_int32

_lib.ml_asr_stream_begin.argtypes = [
    ctypes.POINTER(ml_ASR),
    ctypes.POINTER(ml_AsrStreamBeginInput),
    ctypes.POINTER(ml_AsrStreamBeginOutput),
]
_lib.ml_asr_stream_begin.restype = ctypes.c_int32

_lib.ml_asr_stream_push_audio.argtypes = [ctypes.POINTER(ml_ASR), ctypes.POINTER(ml_AsrStreamPushAudioInput)]
_lib.ml_asr_stream_push_audio.restype = ctypes.c_int32

_lib.ml_asr_stream_stop.argtypes = [ctypes.POINTER(ml_ASR), ctypes.POINTER(ml_AsrStreamStopInput)]
_lib.ml_asr_stream_stop.restype = ctypes.c_int32



import json
import logging
import os
import re
import shutil
from datetime import datetime, timezone
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

from huggingface_hub import HfApi

from .progress_tracker import DownloadProgressTracker
from .types import DownloadProgressInfo, ModelFileInfo, ModelManifest

logger = logging.getLogger(__name__)

# Quant regex pattern matching Go implementation
# Matches: FP32, FP16, F64, F32, F16, I64, I32, I16, I8,
#          Q8_0, Q8_1, Q8_K, Q6_K, Q5_0, Q5_1, Q5_K, Q4_0, Q4_1, Q4_K, Q3_K, Q2_K,
#          IQ4_NL, IQ4_XS, IQ3_S, IQ3_XXS, IQ2_XXS, IQ2_S, IQ2_XS, IQ1_S, IQ1_M,
#          BF16, 1bit, 2bit, 3bit, 4bit, 16bit, etc.
QUANT_REGEX = re.compile(
    r'('
    r'[fF][pP][0-9]+|'  # FP32, FP16, FP64
    r'[fF][0-9]+|'  # F64, F32, F16
    r'[iI][0-9]+|'  # I64, I32, I16, I8
    r'[qQ][0-9]+(_[A-Za-z0-9]+)*|'  # Q8_0, Q8_1, Q8_K, Q6_K, Q5_0, Q5_1, Q5_K, Q4_0, Q4_1, Q4_K, Q3_K, Q2_K
    r'[iI][qQ][0-9]+(_[A-Za-z0-9]+)*|'  # IQ4_NL, IQ4_XS, IQ3_S, IQ3_XXS, IQ2_XXS, IQ2_S, IQ2_XS, IQ1_S, IQ1_M
    r'[bB][fF][0-9]+|'  # BF16
    r'[0-9]+[bB][iI][tT]'  # 1bit, 2bit, 3bit, 4bit, 16bit, 1BIT, 16Bit, etc.
    r')'
)

# Regex to match part files like "-00003-of-00003.gguf"
PART_REGEX = re.compile(r'-\d+-of-\d+\.gguf$')


def get_quant(name: str) -> str:
    match = QUANT_REGEX.search(name)
    if match:
        return match.group(0).upper()
    return 'N/A'


def sum_size(files: List[ModelFileInfo]) -> int:
    return sum(f.size for f in files)


def _update_manifest_file_info(manifest: ModelManifest, filename: str, size: int) -> None:
    # Check all possible locations
    for mf_info in manifest.model_file.values():
        if mf_info.name == filename:
            mf_info.size = size
            mf_info.downloaded = True
            return

    if manifest.mmproj_file.name == filename:
        manifest.mmproj_file.size = size
        manifest.mmproj_file.downloaded = True
        return

    if manifest.tokenizer_file.name == filename:
        manifest.tokenizer_file.size = size
        manifest.tokenizer_file.downloaded = True
        return

    for extra_file in manifest.extra_files:
        if extra_file.name == filename:
            extra_file.size = size
            extra_file.downloaded = True
            return


def choose_files(
    repo_id: str,
    specified_quant: Optional[str],
    files: List[ModelFileInfo],
    existing_manifest: Optional[ModelManifest] = None,
    model_type: Optional[str] = None,
) -> ModelManifest:
    if not files:
        raise ValueError('Repository is empty')

    # Extract model name from repo_id
    model_name = repo_id.split('/')[-1] if '/' in repo_id else repo_id

    # Determine model_type: priority: existing_manifest > passed parameter > default 'llm'
    final_model_type = 'llm'  # Default
    if existing_manifest and existing_manifest.model_type:
        final_model_type = existing_manifest.model_type
    elif model_type:
        final_model_type = model_type

    manifest = ModelManifest(
        name=repo_id,
        model_name=model_name,
        model_type=final_model_type,
    )

    # Copy metadata from existing manifest if available
    if existing_manifest:
        manifest.model_name = existing_manifest.model_name
        manifest.plugin_id = existing_manifest.plugin_id
        manifest.device_id = existing_manifest.device_id
        manifest.model_type = existing_manifest.model_type  # Keep from manifest if exists
        manifest.min_sdk_version = existing_manifest.min_sdk_version

    # Classify files
    mmprojs: List[ModelFileInfo] = []
    tokenizers: List[ModelFileInfo] = []
    onnx_files: List[ModelFileInfo] = []
    nexa_files: List[ModelFileInfo] = []
    npy_files: List[ModelFileInfo] = []
    ggufs: Dict[str, List[ModelFileInfo]] = {}  # key is gguf name without part

    for file_info in files:
        name_lower = file_info.name.lower()
        if name_lower.endswith('.gguf'):
            if name_lower.startswith('mmproj'):
                mmprojs.append(file_info)
            else:
                # Remove part suffix (e.g., "-00003-of-00003.gguf")
                base_name = PART_REGEX.sub('', file_info.name)
                if base_name not in ggufs:
                    ggufs[base_name] = []
                ggufs[base_name].append(file_info)
        elif name_lower.endswith('tokenizer.json'):
            tokenizers.append(file_info)
        elif name_lower.endswith('.onnx'):
            onnx_files.append(file_info)
        elif name_lower.endswith('.nexa'):
            nexa_files.append(file_info)
        elif name_lower.endswith('.npy'):
            npy_files.append(file_info)

    # Choose model files
    if ggufs:
        # Handle GGUF files
        if len(ggufs) == 1:
            # Single quant
            base_name, gguf_list = next(iter(ggufs.items()))
            quant_name = get_quant(base_name) or 'default'
            gguf_list_sorted = sorted(gguf_list, key=lambda f: f.name)

            manifest.model_file[quant_name] = ModelFileInfo(
                name=gguf_list_sorted[0].name,
                downloaded=True,
                size=sum_size(gguf_list),
            )
            # Add other fragments as extra files
            manifest.extra_files.extend(
                ModelFileInfo(name=f.name, downloaded=True, size=f.size) for f in gguf_list_sorted[1:]
            )

            if specified_quant and specified_quant not in manifest.model_file:
                raise ValueError(f'Specified quant {specified_quant} not found')
        else:
            # Multiple quants - choose one
            gguf_names = list(ggufs.keys())

            # Sort by size (largest first)
            gguf_names.sort(key=lambda k: sum_size(ggufs[k]), reverse=True)

            # Find default file (prefer Q8_0, Q4_K_M, Q4_0)
            default_file = None
            preferred_quants = ['Q8_0', 'Q4_K_M', 'Q4_0']
            for gguf_name in gguf_names:
                quant = get_quant(gguf_name)
                if quant in preferred_quants:
                    default_quant = get_quant(default_file) if default_file else None
                    if default_file is None or preferred_quants.index(quant) < preferred_quants.index(default_quant):
                        default_file = gguf_name

            if default_file is None:
                default_file = gguf_names[0]

            # If specified quant, find matching file
            selected_file = default_file
            if specified_quant:
                selected_file = None
                for gguf_name in gguf_names:
                    if get_quant(gguf_name).upper() == specified_quant.upper():
                        selected_file = gguf_name
                        break
                if selected_file is None:
                    available_quants = [get_quant(name) for name in gguf_names]
                    raise ValueError(
                        f'Specified quant {specified_quant} not found. Available quants: {available_quants}'
                    )

            # Create manifest entries for all quants
            for gguf_name, gguf_list in ggufs.items():
                quant_name = get_quant(gguf_name) or 'default'
                gguf_list_sorted = sorted(gguf_list, key=lambda f: f.name)
                downloaded = gguf_name == selected_file

                manifest.model_file[quant_name] = ModelFileInfo(
                    name=gguf_list_sorted[0].name,
                    downloaded=downloaded,
                    size=sum_size(gguf_list),
                )
                # Add other fragments as extra files
                manifest.extra_files.extend(
                    ModelFileInfo(name=f.name, downloaded=downloaded, size=f.size) for f in gguf_list_sorted[1:]
                )

        # Handle mmproj
        if not mmprojs:
            # Fallback to onnx or nexa file as mmproj
            if len(onnx_files) == 1:
                manifest.mmproj_file = ModelFileInfo(name=onnx_files[0].name, downloaded=True, size=onnx_files[0].size)
            elif len(nexa_files) == 1:
                manifest.mmproj_file = ModelFileInfo(name=nexa_files[0].name, downloaded=True, size=nexa_files[0].size)
        elif len(mmprojs) == 1:
            manifest.mmproj_file = ModelFileInfo(name=mmprojs[0].name, downloaded=True, size=mmprojs[0].size)
        else:
            # Choose biggest
            biggest = max(mmprojs, key=lambda f: f.size)
            manifest.mmproj_file = ModelFileInfo(name=biggest.name, downloaded=True, size=biggest.size)

        # Handle tokenizer
        if len(tokenizers) == 1:
            manifest.tokenizer_file = ModelFileInfo(
                name=tokenizers[0].name,
                downloaded=True,
                size=tokenizers[0].size,
            )
        elif len(tokenizers) > 1:
            raise ValueError(
                f'Multiple tokenizer files found: {[t.name for t in tokenizers]}. Expected exactly one tokenizer file'
            )

        # Add .nexa files as extra files (except if used as mmproj)
        manifest.extra_files.extend(
            ModelFileInfo(name=f.name, downloaded=True, size=f.size)
            for f in nexa_files
            if manifest.mmproj_file.name != f.name
        )

        # Add .npy files as extra files
        manifest.extra_files.extend(ModelFileInfo(name=f.name, downloaded=True, size=f.size) for f in npy_files)
    else:
        # Non-GGUF models (MLX, etc.)
        if specified_quant:
            raise ValueError(f'Specified quant {specified_quant} only supported in GGUF models')

        # Extract quant from repo name
        quant = get_quant(repo_id)

        # Find main model file (prefer non-nested files)
        supported_extensions = ['.safetensors', '.npz', '.nexa', '.bin']
        main_file = None
        for file_info in files:
            name_lower = file_info.name.lower()
            if any(name_lower.endswith(ext) for ext in supported_extensions):
                if '/' not in file_info.name:
                    main_file = file_info
                    break

        # Fallback to any supported file, then first file
        if main_file is None:
            for file_info in files:
                name_lower = file_info.name.lower()
                if any(name_lower.endswith(ext) for ext in supported_extensions):
                    main_file = file_info
                    break
            if main_file is None:
                main_file = files[0]

        manifest.model_file[quant] = ModelFileInfo(
            name=main_file.name,
            downloaded=True,
            size=main_file.size,
        )

        # Add other files as extra files
        manifest.extra_files.extend(
            ModelFileInfo(name=f.name, downloaded=True, size=f.size) for f in files if f.name != main_file.name
        )

    # Set default values
    if not manifest.model_name:
        manifest.model_name = repo_id.split('/')[-1] if '/' in repo_id else repo_id

    if not manifest.plugin_id:
        manifest.plugin_id = 'metal' if 'mlx' in repo_id.lower() else 'cpu_gpu'

    return manifest


def is_local_path(path: str) -> bool:
    if os.path.isabs(path) or os.path.exists(path) or '\\' in path:
        return True
    if '/' in path:
        if path.count('/') == 1:
            return path.startswith('./') or path.startswith('../') or os.path.exists(path)
        return True
    return os.path.exists(path)


def get_model_info(
    repo_id: str,
    token: Optional[str] = None,
    custom_endpoint: Optional[str] = None,
) -> Tuple[List[ModelFileInfo], Optional[ModelManifest], Optional[str]]:
    api = HfApi(token=token, endpoint=custom_endpoint) if custom_endpoint else HfApi(token=token)

    try:
        files = api.list_repo_files(repo_id=repo_id, repo_type='model')
    except Exception as e:
        logger.error(f'Failed to list files for {repo_id}: {e}')
        raise

    file_size_map = {}
    pipeline_tag = None
    try:
        repo_info = api.repo_info(repo_id=repo_id, repo_type='model', files_metadata=True)
        if repo_info:
            # Get pipeline_tag from model info
            if hasattr(repo_info, 'pipeline_tag') and repo_info.pipeline_tag:
                pipeline_tag = repo_info.pipeline_tag
            # Get file sizes
            if hasattr(repo_info, 'siblings') and repo_info.siblings:
                for sibling in repo_info.siblings:
                    if hasattr(sibling, 'rfilename') and hasattr(sibling, 'size'):
                        file_size_map[sibling.rfilename] = sibling.size
    except Exception as e:
        logger.warning(f'Failed to get repo info for {repo_id}: {e}')

    file_infos = [ModelFileInfo(name=f, downloaded=False, size=file_size_map.get(f, 0)) for f in files]
    has_manifest = 'nexa.manifest' in files

    manifest = None
    if has_manifest:
        try:
            manifest_path = api.hf_hub_download(repo_id=repo_id, filename='nexa.manifest', repo_type='model')
            with open(manifest_path, 'r', encoding='utf-8') as f:
                manifest = ModelManifest.from_dict(json.load(f))
        except Exception as e:
            logger.warning(f'Failed to download manifest for {repo_id}: {e}')

    return file_infos, manifest, pipeline_tag


def download_model(
    repo_id: str,
    output_dir: Path,
    files: List[ModelFileInfo],
    token: Optional[str] = None,
    quant_spec: Optional[str] = None,
    existing_manifest: Optional[ModelManifest] = None,
    progress_callback: Optional[Callable[[DownloadProgressInfo], None]] = None,
    enable_transfer: bool = True,
    show_progress: bool = True,
    custom_endpoint: Optional[str] = None,
    force_download: bool = False,
    model_type: Optional[str] = None,
    **kwargs,
) -> ModelManifest:
    output_dir.mkdir(parents=True, exist_ok=True)
    manifest = choose_files(repo_id, quant_spec, files, existing_manifest, model_type=model_type)

    files_to_download = [f for f in manifest.model_file.values() if f.downloaded and f.name]
    if manifest.mmproj_file.downloaded and manifest.mmproj_file.name:
        files_to_download.append(manifest.mmproj_file)
    if manifest.tokenizer_file.downloaded and manifest.tokenizer_file.name:
        files_to_download.append(manifest.tokenizer_file)
    files_to_download.extend(f for f in manifest.extra_files if f.downloaded and f.name)

    logger.debug(f'Downloading {len(files_to_download)} files from {repo_id}...')

    os.environ['HF_HUB_ENABLE_HF_TRANSFER'] = '1' if enable_transfer else '0'
    hf_token = token or os.getenv('HF_TOKEN')
    api = HfApi(token=hf_token, endpoint=custom_endpoint) if custom_endpoint else HfApi(token=hf_token)
    total_size = sum(f.size for f in files_to_download)
    file_count = len(files_to_download)

    tracker: Optional[DownloadProgressTracker] = None
    if progress_callback:
        tracker = DownloadProgressTracker(
            progress_callback=progress_callback,
            show_progress=show_progress,
        )
        tracker.set_repo_info(total_size, file_count)
        tracker.start_tracking()

    try:
        for file_info in files_to_download:
            try:
                expected_path = output_dir / file_info.name

                # Skip existing files unless force_download is True
                if not force_download and expected_path.exists() and file_info.size > 0:
                    actual_size = expected_path.stat().st_size
                    if actual_size == file_info.size:
                        logger.debug(f'File already exists: {file_info.name} (size: {actual_size})')
                        _update_manifest_file_info(manifest, file_info.name, actual_size)
                        continue

                path = api.hf_hub_download(
                    repo_id=repo_id,
                    filename=file_info.name,
                    local_dir=str(output_dir),
                    force_download=force_download,
                )
                if path and os.path.exists(path):
                    expected_path.parent.mkdir(parents=True, exist_ok=True)
                    if os.path.isdir(path) and not expected_path.exists():
                        shutil.copytree(path, expected_path, dirs_exist_ok=True)
                    elif path != str(expected_path):
                        shutil.copy2(path, expected_path)

                actual_size = expected_path.stat().st_size if expected_path.exists() else 0
                _update_manifest_file_info(manifest, file_info.name, actual_size)
                logger.debug(f'Downloaded: {file_info.name} to {expected_path} (size: {actual_size})')
            except Exception as e:
                logger.warning(f'Failed to download {file_info.name}: {e}')
                if tracker:
                    tracker.set_error(f'Failed to download {file_info.name}: {e}')
    finally:
        if tracker:
            tracker.stop_tracking()

    manifest.pipeline_tag = kwargs.get('pipeline_tag')
    manifest.avatar_url = kwargs.get('avatar_url')
    manifest.download_time = datetime.now(timezone.utc).isoformat()
    return manifest


def ensure_model_downloaded(
    repo_id: str,
    quant_spec: Optional[str] = None,
    token: Optional[str] = None,
    progress_callback: Optional[Callable[[DownloadProgressInfo], None]] = None,
    local_dir: Optional[str] = None,
    enable_transfer: bool = True,
    show_progress: bool = True,
    custom_endpoint: Optional[str] = None,
    force_download: bool = False,
    model_type: Optional[str] = None,
    **kwargs,
) -> 'ModelManifest':
    from .store import Store

    store = Store.get()
    model_dir = Path(local_dir) / repo_id if local_dir else store.model_dir_path() / repo_id
    manifest = store.get_manifest(repo_id) if not force_download and store.get_manifest(repo_id) else None
    if manifest:
        logger.debug(f'Model {repo_id} already exists locally')
        return manifest

    logger.debug(f'Model {repo_id} not found locally, downloading from Huggingface...')
    if not store.lock_model(repo_id):
        raise RuntimeError(f'Failed to acquire lock for model {repo_id}')

    try:
        files, existing_manifest, pipeline_tag = get_model_info(repo_id, token=token, custom_endpoint=custom_endpoint)
        manifest = download_model(
            repo_id=repo_id,
            output_dir=model_dir,
            files=files,
            token=token,
            quant_spec=quant_spec,
            existing_manifest=existing_manifest,
            progress_callback=progress_callback,
            pipeline_tag=pipeline_tag,
            model_type=model_type,
            enable_transfer=enable_transfer,
            show_progress=show_progress,
            custom_endpoint=custom_endpoint,
            force_download=force_download,
            **kwargs,
        )
        store.save_manifest(manifest)
    finally:
        store.unlock_model(repo_id)

    return manifest



"""
TTS (Text-to-Speech) wrapper for nexaai.
Provides high-level Python API for TTS operations.
"""

import ctypes
import logging
from typing import List, Optional

from .internal.model_loader import ModelLoaderMixin
from .internal.native_plugin import (
    get_mlx_tts_class,
    is_native_plugin,
    setup_mlx_imports,
)
from .nexa_sdk._lib import _lib
from .nexa_sdk.error import check_error
from .nexa_sdk.tts import (
    ml_TTS,
    ml_TTSConfig,
    ml_TtsCreateInput,
    ml_TtsListAvailableVoicesInput,
    ml_TtsListAvailableVoicesOutput,
    ml_TTSResult,
    ml_TtsSynthesizeInput,
    ml_TtsSynthesizeOutput,
)
from .nexa_sdk.types import ModelConfig, ProfileData

logger = logging.getLogger(__name__)


class TTS(ModelLoaderMixin):
    """Text-to-Speech wrapper."""

    def __init__(
        self,
        model_path: str,
        model_name: Optional[str] = None,
        config: Optional[ModelConfig] = None,
        vocoder_path: Optional[str] = None,
        plugin_id: Optional[str] = None,
        device_id: Optional[str] = None,
    ):
        """
        Create and initialize a TTS instance.

        Args:
            model_path: Path to the model file.
            model_name: Name of the model. If None, uses model_path.
            config: Model configuration. If None, uses default config.
            vocoder_path: Path to the vocoder file. If None, uses default.
            plugin_id: Plugin to use for the model. If None, uses default.
            device_id: Device to use for the model. If None, uses default device.
        """
        if config is None:
            config = ModelConfig()

        if model_name is None:
            model_name = model_path

        if not model_name:
            model_name = model_path

        # Check if we should use native Python calls (for MLX plugin)
        self._is_native = False
        is_native = is_native_plugin(plugin_id)
        logger.debug(f'is_native_plugin({plugin_id}) = {is_native}')
        if is_native:
            logger.info(f'Plugin {plugin_id} is a native plugin, attempting native initialization')
            imports_ok = setup_mlx_imports()
            logger.info(f'setup_mlx_imports() = {imports_ok}')
            if not imports_ok:
                raise RuntimeError(
                    f'Failed to setup MLX imports for native plugin {plugin_id}. '
                    f'Cannot fallback to C interface as it would cause GIL conflicts.'
                )
            try:
                logger.info('Initializing MLX native TTS')
                self._init_mlx_native(
                    model_path=model_path,
                    vocoder_path=vocoder_path,
                    device_id=device_id,
                )
                logger.info('MLX native TTS initialized successfully')
                self._config = config
                return  # Native initialization successful, skip C interface
            except Exception as e:
                raise RuntimeError(
                    f'Failed to initialize MLX native for plugin {plugin_id}: {e}. '
                    f'Cannot fallback to C interface as it would cause GIL conflicts.'
                ) from e
        else:
            logger.debug(f'Plugin {plugin_id} is not a native plugin, using C interface')

        # C interface initialization (default or fallback)
        c_config = config.to_c_struct()

        model_name_bytes = model_name.encode('utf-8')
        model_path_bytes = model_path.encode('utf-8')
        vocoder_path_bytes = vocoder_path.encode('utf-8') if vocoder_path else None
        plugin_id_str = plugin_id if plugin_id else ''
        plugin_id_buf = ctypes.create_string_buffer(plugin_id_str.encode('utf-8'))
        device_id_buf = ctypes.create_string_buffer(device_id.encode('utf-8')) if device_id else None

        model_name_buf = ctypes.create_string_buffer(model_name_bytes)
        model_path_buf = ctypes.create_string_buffer(model_path_bytes)
        vocoder_path_buf = ctypes.create_string_buffer(vocoder_path_bytes) if vocoder_path_bytes else None

        c_input = ml_TtsCreateInput(
            model_name=ctypes.cast(model_name_buf, ctypes.c_char_p),
            model_path=ctypes.cast(model_path_buf, ctypes.c_char_p),
            config=c_config,
            vocoder_path=ctypes.cast(vocoder_path_buf, ctypes.c_char_p) if vocoder_path_buf else None,
            plugin_id=ctypes.cast(plugin_id_buf, ctypes.c_char_p) if plugin_id_buf else None,
            device_id=ctypes.cast(device_id_buf, ctypes.c_char_p) if device_id_buf else None,
        )

        self._string_refs = [
            model_name_buf,
            model_path_buf,
            vocoder_path_buf,
            plugin_id_buf,
            device_id_buf,
        ]

        handle_ptr = ctypes.POINTER(ml_TTS)()
        error_code = _lib.ml_tts_create(ctypes.pointer(c_input), ctypes.byref(handle_ptr))
        check_error(error_code)

        if not handle_ptr:
            raise RuntimeError('Failed to create TTS instance')

        self._handle = handle_ptr
        self._config = config

    def _init_mlx_native(
        self,
        model_path: str,
        vocoder_path: Optional[str],
        device_id: Optional[str],
    ):
        """
        Initialize TTS using MLX native Python implementation.

        Args:
            model_path: Path to the model file.
            vocoder_path: Path to the vocoder file.
            device_id: Device ID (e.g., 'cpu', 'gpu').
        """
        MLXTTS = get_mlx_tts_class()

        # Use model_path as vocoder_path if vocoder_path is not provided
        vocoder_path_str = vocoder_path if vocoder_path else model_path

        # Create MLX TTS instance
        self._mlx_tts = MLXTTS(
            model_path=model_path,
            vocoder_path=vocoder_path_str,
            device=device_id if device_id else 'cpu',
        )

        self._is_native = True
        self._handle = None  # No C handle for native implementation

    def __del__(self):
        """Destroy TTS instance and free associated resources."""
        if hasattr(self, '_is_native') and self._is_native:
            # Native implementation cleanup
            if hasattr(self, '_mlx_tts') and self._mlx_tts:
                try:
                    self._mlx_tts.destroy()
                except Exception as e:
                    logger.warning(f'Error during MLX native TTS cleanup: {e}')
                self._mlx_tts = None
        elif hasattr(self, '_handle') and self._handle:
            # C interface cleanup
            try:
                _lib.ml_tts_destroy(self._handle)
            except Exception as e:
                logger.warning(f'Error during TTS cleanup: {e}')
            self._handle = None

    @classmethod
    def _get_model_type(cls) -> str:
        return 'tts'

    @classmethod
    def _extract_model_params_from_manifest(
        cls,
        manifest,
        model_path: str,
        repo_id: str,
        store,
        **kwargs,
    ) -> dict:
        """Extract TTS-specific parameters from manifest."""
        model_name = manifest.model_name or repo_id
        if not model_name:
            model_name = repo_id

        vocoder_path = None
        if hasattr(manifest, 'vocoder_file') and manifest.vocoder_file and manifest.vocoder_file.name:
            vocoder_path_obj = store.modelfile_path(repo_id, manifest.vocoder_file.name)
            if vocoder_path_obj.exists():
                vocoder_path = str(vocoder_path_obj)

        plugin_id = manifest.plugin_id.strip() if manifest.plugin_id else None
        if not plugin_id:
            plugin_id = kwargs.pop('plugin_id', None)
        else:
            # Remove plugin_id from kwargs if manifest has it (to avoid passing it twice)
            kwargs.pop('plugin_id', None)
        if plugin_id:
            plugin_id = plugin_id.strip() if isinstance(plugin_id, str) else plugin_id
            if not plugin_id:
                plugin_id = None

        if not plugin_id:
            plugin_id = cls._get_default_plugin_id()

        device_id = manifest.device_id.strip() if manifest.device_id else None
        if not device_id:
            device_id = kwargs.pop('device_id', None)
        if device_id:
            device_id = device_id.strip() if isinstance(device_id, str) else device_id
            if not device_id:
                device_id = None

        return {
            'model_path': str(model_path),
            'model_name': model_name,
            'vocoder_path': vocoder_path,
            'plugin_id': plugin_id,
            'device_id': device_id,
            **kwargs,
        }

    @classmethod
    def _create_instance(cls, **params) -> 'TTS':
        """Create TTS instance with given parameters."""
        return cls(**params)

    @classmethod
    def _get_default_plugin_id(cls) -> str:
        """Get default plugin ID for TTS."""
        return 'metal'

    def synthesize(
        self,
        text: str,
        output_path: str,
        voice: Optional[str] = None,
        speed: float = 1.0,
        seed: int = -1,
        sample_rate: int = 22050,
    ) -> 'SynthesizeResult':
        """
        Synthesize text to speech.

        Args:
            text: Text to synthesize.
            output_path: Path to save the audio file.
            voice: Voice identifier. If None, uses default.
            speed: Speech speed multiplier.
            seed: Random seed. -1 for random.
            sample_rate: Audio sample rate.

        Returns:
            SynthesizeResult containing audio file path and metadata.

        Raises:
            NexaError: If synthesis fails.
        """
        if hasattr(self, '_is_native') and self._is_native:
            # Native implementation
            from ml import TTSConfig as MLXTTSConfig

            mlx_config = MLXTTSConfig(
                voice=voice or 'default',
                speed=speed,
                seed=seed,
                sample_rate=sample_rate,
            )

            mlx_result = self._mlx_tts.synthesize(
                text=text,
                config=mlx_config,
                output_path=output_path,
                clear_cache=True,
            )

            # Convert MLX result to SynthesizeResult
            from .nexa_sdk.types import ProfileData

            return SynthesizeResult(
                audio_path=mlx_result.audio_path,
                duration_seconds=mlx_result.duration_seconds,
                sample_rate=mlx_result.sample_rate,
                channels=mlx_result.channels,
                num_samples=mlx_result.num_samples,
                profile_data=ProfileData(),  # MLX may not provide profile data
            )

        # C interface implementation
        tts_config = ml_TTSConfig(
            voice=voice.encode('utf-8') if voice else None,
            speed=ctypes.c_float(speed),
            seed=ctypes.c_int32(seed),
            sample_rate=ctypes.c_int32(sample_rate),
        )

        text_bytes = text.encode('utf-8')
        text_buf = ctypes.create_string_buffer(text_bytes)
        output_path_bytes = output_path.encode('utf-8')
        output_path_buf = ctypes.create_string_buffer(output_path_bytes)

        c_input = ml_TtsSynthesizeInput(
            text_utf8=ctypes.cast(text_buf, ctypes.c_char_p),
            config=ctypes.pointer(tts_config),
            output_path=ctypes.cast(output_path_buf, ctypes.c_char_p),
        )

        c_output = ml_TtsSynthesizeOutput()
        error_code = _lib.ml_tts_synthesize(self._handle, ctypes.pointer(c_input), ctypes.pointer(c_output))
        check_error(error_code)

        return self._extract_result(c_output)

    def _extract_result(self, c_output: ml_TtsSynthesizeOutput) -> 'SynthesizeResult':
        """Extract SynthesizeResult from C output structure."""
        result = c_output.result
        audio_path = ''
        if result.audio_path:
            audio_path = result.audio_path.decode('utf-8')

        profile_data = ProfileData.from_c_struct(c_output.profile_data)
        return SynthesizeResult(
            audio_path=audio_path,
            duration_seconds=result.duration_seconds,
            sample_rate=result.sample_rate,
            channels=result.channels,
            num_samples=result.num_samples,
            profile_data=profile_data,
        )

    def list_available_voices(self) -> List[str]:
        """
        List available voices.

        Returns:
            List of voice identifiers.

        Raises:
            NexaError: If query fails.
        """
        if hasattr(self, '_is_native') and self._is_native:
            # Native implementation
            return self._mlx_tts.list_available_voices()

        # C interface implementation
        c_input = ml_TtsListAvailableVoicesInput(reserved=None)
        c_output = ml_TtsListAvailableVoicesOutput()

        error_code = _lib.ml_tts_list_available_voices(
            self._handle,
            ctypes.pointer(c_input),
            ctypes.pointer(c_output),
        )
        check_error(error_code)

        voices = []
        if c_output.voice_ids and c_output.voice_count > 0:
            for i in range(c_output.voice_count):
                voice_ptr = c_output.voice_ids[i]
                if voice_ptr:
                    voices.append(voice_ptr.decode('utf-8'))
            _lib.ml_free(c_output.voice_ids)

        return voices


class SynthesizeResult:
    """Result of TTS synthesis."""

    def __init__(
        self,
        audio_path: str,
        duration_seconds: float,
        sample_rate: int,
        channels: int,
        num_samples: int,
        profile_data: ProfileData,
    ):
        self.audio_path = audio_path
        self.duration_seconds = duration_seconds
        self.sample_rate = sample_rate
        self.channels = channels
        self.num_samples = num_samples
        self.profile_data = profile_data

    def __repr__(self) -> str:
        return f'SynthesizeResult(audio_path={self.audio_path}, duration={self.duration_seconds}s, sample_rate={self.sample_rate})'



"""
Embedding functions and structures for nexa-sdk.
"""

import ctypes

from ._lib import _lib
from .error import ml_log_callback
from .llm import ml_ModelConfig, ml_ProfileData

# ============================================================================
# C Structure Definitions
# ============================================================================


class ml_EmbeddingConfig(ctypes.Structure):
    """Embedding generation configuration."""

    _fields_ = [
        ('batch_size', ctypes.c_int32),
        ('normalize', ctypes.c_bool),
        ('normalize_method', ctypes.c_char_p),
    ]


class ml_Embedder(ctypes.Structure):
    """Opaque embedder handle."""

    pass


class ml_EmbedderCreateInput(ctypes.Structure):
    """Input structure for creating an embedder."""

    _fields_ = [
        ('model_name', ctypes.c_char_p),
        ('model_path', ctypes.c_char_p),
        ('tokenizer_path', ctypes.c_char_p),
        ('config', ml_ModelConfig),
        ('plugin_id', ctypes.c_char_p),
        ('device_id', ctypes.c_char_p),
    ]


class ml_EmbedderEmbedInput(ctypes.Structure):
    """Input structure for embedding generation."""

    _fields_ = [
        ('texts', ctypes.POINTER(ctypes.c_char_p)),
        ('text_count', ctypes.c_int32),
        ('config', ctypes.POINTER(ml_EmbeddingConfig)),
        ('input_ids_2d', ctypes.POINTER(ctypes.POINTER(ctypes.c_int32))),
        ('input_ids_row_lengths', ctypes.POINTER(ctypes.c_int32)),
        ('input_ids_row_count', ctypes.c_int32),
        ('task_type', ctypes.c_char_p),
        ('image_paths', ctypes.POINTER(ctypes.c_char_p)),
        ('image_count', ctypes.c_int32),
    ]


class ml_EmbedderEmbedOutput(ctypes.Structure):
    """Output structure for embedding generation."""

    _fields_ = [
        ('embeddings', ctypes.POINTER(ctypes.c_float)),
        ('embedding_count', ctypes.c_int32),
        ('profile_data', ml_ProfileData),
    ]


class ml_EmbedderDimOutput(ctypes.Structure):
    """Output structure for getting embedding dimension."""

    _fields_ = [
        ('dimension', ctypes.c_int32),
    ]


# ============================================================================
# Function Signatures
# ============================================================================

_lib.ml_embedder_create.argtypes = [
    ctypes.POINTER(ml_EmbedderCreateInput),
    ctypes.POINTER(ctypes.POINTER(ml_Embedder)),
]
_lib.ml_embedder_create.restype = ctypes.c_int32

_lib.ml_embedder_destroy.argtypes = [ctypes.POINTER(ml_Embedder)]
_lib.ml_embedder_destroy.restype = ctypes.c_int32

_lib.ml_embedder_embed.argtypes = [
    ctypes.POINTER(ml_Embedder),
    ctypes.POINTER(ml_EmbedderEmbedInput),
    ctypes.POINTER(ml_EmbedderEmbedOutput),
]
_lib.ml_embedder_embed.restype = ctypes.c_int32

_lib.ml_embedder_embedding_dim.argtypes = [ctypes.POINTER(ml_Embedder), ctypes.POINTER(ml_EmbedderDimOutput)]
_lib.ml_embedder_embedding_dim.restype = ctypes.c_int32



"""
CV (Computer Vision) wrapper for nexaai.
Provides high-level Python API for computer vision operations.
"""

import ctypes
import logging
from typing import List, Optional

from .internal.model_loader import ModelLoaderMixin
from .internal.native_plugin import (
    get_mlx_cv_class,
    is_native_plugin,
    setup_mlx_imports,
)
from .nexa_sdk._lib import _lib
from .nexa_sdk.cv import (
    ml_CV,
    ml_CVCapabilities,
    ml_CVCreateInput,
    ml_CVInferInput,
    ml_CVInferOutput,
    ml_CVModelConfig,
)
from .nexa_sdk.error import check_error

logger = logging.getLogger(__name__)


class CV(ModelLoaderMixin):
    """Computer Vision model wrapper."""

    def __init__(
        self,
        model_name: str,
        capabilities: int = 0,
        det_model_path: Optional[str] = None,
        rec_model_path: Optional[str] = None,
        char_dict_path: Optional[str] = None,
        qnn_model_folder_path: Optional[str] = None,
        qnn_lib_folder_path: Optional[str] = None,
        plugin_id: Optional[str] = None,
        device_id: Optional[str] = None,
        license_id: Optional[str] = None,
        license_key: Optional[str] = None,
    ):
        """
        Create and initialize a CV instance.

        Args:
            model_name: Name of the model.
            capabilities: CV capabilities (0=OCR, 1=Classification, 2=Segmentation, 3=Custom).
            det_model_path: Path to detection model (for OCR).
            rec_model_path: Path to recognition model (for OCR).
            char_dict_path: Path to character dictionary (for OCR).
            qnn_model_folder_path: Path to QNN model folder.
            qnn_lib_folder_path: Path to QNN library folder.
            plugin_id: Plugin to use for the model. If None, uses default.
            device_id: Device to use for the model. If None, uses default device.
            license_id: License ID for loading NPU models.
            license_key: License key for loading NPU models.
        """
        # Check if we should use native Python calls (for MLX plugin)
        self._is_native = False
        is_native = is_native_plugin(plugin_id)
        logger.debug(f'is_native_plugin({plugin_id}) = {is_native}')
        if is_native:
            logger.info(f'Plugin {plugin_id} is a native plugin, attempting native initialization')
            imports_ok = setup_mlx_imports()
            logger.info(f'setup_mlx_imports() = {imports_ok}')
            if not imports_ok:
                raise RuntimeError(
                    f'Failed to setup MLX imports for native plugin {plugin_id}. '
                    f'Cannot fallback to C interface as it would cause GIL conflicts.'
                )
            try:
                logger.info('Initializing MLX native CV')
                self._init_mlx_native(
                    capabilities=capabilities,
                    det_model_path=det_model_path,
                    rec_model_path=rec_model_path,
                    device_id=device_id,
                )
                logger.info('MLX native CV initialized successfully')
                return  # Native initialization successful, skip C interface
            except Exception as e:
                raise RuntimeError(
                    f'Failed to initialize MLX native for plugin {plugin_id}: {e}. '
                    f'Cannot fallback to C interface as it would cause GIL conflicts.'
                ) from e
        else:
            logger.debug(f'Plugin {plugin_id} is not a native plugin, using C interface')

        model_name_bytes = model_name.encode('utf-8')
        plugin_id_str = plugin_id if plugin_id else ''
        device_id_str = device_id if device_id else ''
        license_id_str = license_id if license_id else ''
        license_key_str = license_key if license_key else ''

        model_name_buf = ctypes.create_string_buffer(model_name_bytes)
        plugin_id_buf = ctypes.create_string_buffer(plugin_id_str.encode('utf-8'))
        device_id_buf = ctypes.create_string_buffer(device_id_str.encode('utf-8'))
        license_id_buf = ctypes.create_string_buffer(license_id_str.encode('utf-8'))
        license_key_buf = ctypes.create_string_buffer(license_key_str.encode('utf-8'))

        det_model_path_buf = ctypes.create_string_buffer(det_model_path.encode('utf-8')) if det_model_path else None
        rec_model_path_buf = ctypes.create_string_buffer(rec_model_path.encode('utf-8')) if rec_model_path else None
        char_dict_path_buf = ctypes.create_string_buffer(char_dict_path.encode('utf-8')) if char_dict_path else None
        qnn_model_folder_path_buf = (
            ctypes.create_string_buffer(qnn_model_folder_path.encode('utf-8')) if qnn_model_folder_path else None
        )
        qnn_lib_folder_path_buf = (
            ctypes.create_string_buffer(qnn_lib_folder_path.encode('utf-8')) if qnn_lib_folder_path else None
        )

        # Create cv_model_config with pointers to the string buffers
        cv_model_config = ml_CVModelConfig(
            capabilities=ml_CVCapabilities(capabilities),
            det_model_path=ctypes.cast(det_model_path_buf, ctypes.c_char_p) if det_model_path_buf else None,
            rec_model_path=ctypes.cast(rec_model_path_buf, ctypes.c_char_p) if rec_model_path_buf else None,
            char_dict_path=ctypes.cast(char_dict_path_buf, ctypes.c_char_p) if char_dict_path_buf else None,
            qnn_model_folder_path=(
                ctypes.cast(qnn_model_folder_path_buf, ctypes.c_char_p) if qnn_model_folder_path_buf else None
            ),
            qnn_lib_folder_path=(
                ctypes.cast(qnn_lib_folder_path_buf, ctypes.c_char_p) if qnn_lib_folder_path_buf else None
            ),
        )

        c_input = ml_CVCreateInput(
            model_name=ctypes.cast(model_name_buf, ctypes.c_char_p),
            config=cv_model_config,
            plugin_id=ctypes.cast(plugin_id_buf, ctypes.c_char_p) if plugin_id_buf else None,
            device_id=ctypes.cast(device_id_buf, ctypes.c_char_p) if device_id_buf else None,
            license_id=ctypes.cast(license_id_buf, ctypes.c_char_p) if license_id_buf else None,
            license_key=ctypes.cast(license_key_buf, ctypes.c_char_p) if license_key_buf else None,
        )

        self._string_refs = [
            model_name_buf,
            plugin_id_buf,
            device_id_buf,
            license_id_buf,
            license_key_buf,
            det_model_path_buf,
            rec_model_path_buf,
            char_dict_path_buf,
            qnn_model_folder_path_buf,
            qnn_lib_folder_path_buf,
        ]

        handle_ptr = ctypes.POINTER(ml_CV)()
        error_code = _lib.ml_cv_create(ctypes.pointer(c_input), ctypes.byref(handle_ptr))
        check_error(error_code)

        if not handle_ptr:
            raise RuntimeError('Failed to create CV instance')

        self._handle = handle_ptr

    def _init_mlx_native(
        self,
        capabilities: int,
        det_model_path: Optional[str],
        rec_model_path: Optional[str],
        device_id: Optional[str],
    ):
        """
        Initialize CV using MLX native Python implementation.

        Args:
            capabilities: CV capabilities (0=OCR, 1=Classification, 2=Segmentation, 3=Custom).
            det_model_path: Path to detection model (for OCR).
            rec_model_path: Path to recognition model (for OCR).
            device_id: Device ID (e.g., 'cpu', 'gpu').
        """
        MLXCV = get_mlx_cv_class()

        # Import CVModelConfig from ml
        from ml import CVModelConfig as MLXCVModelConfig

        # Create MLX CV config
        mlx_config = MLXCVModelConfig(
            capabilities=capabilities,
            det_model_path=det_model_path,
            rec_model_path=rec_model_path,
        )

        # Create MLX CV instance
        self._mlx_cv = MLXCV(
            config=mlx_config,
            device=device_id if device_id else 'cpu',
        )

        self._is_native = True
        self._handle = None  # No C handle for native implementation

    def __del__(self):
        """Destroy CV instance and free associated resources."""
        if hasattr(self, '_is_native') and self._is_native:
            # Native implementation cleanup
            if hasattr(self, '_mlx_cv') and self._mlx_cv:
                try:
                    self._mlx_cv.destroy()
                except Exception as e:
                    logger.warning(f'Error during MLX native CV cleanup: {e}')
                self._mlx_cv = None
        elif hasattr(self, '_handle') and self._handle:
            # C interface cleanup
            try:
                _lib.ml_cv_destroy(self._handle)
            except Exception as e:
                logger.warning(f'Error during CV cleanup: {e}')
            self._handle = None

    @classmethod
    def _get_model_type(cls) -> str:
        return 'cv'

    @classmethod
    def _extract_model_params_from_manifest(
        cls,
        manifest,
        model_path: str,
        repo_id: str,
        store,
        **kwargs,
    ) -> dict:
        """Extract CV-specific parameters from manifest."""
        model_name = manifest.model_name or repo_id
        if not model_name:
            model_name = repo_id

        capabilities = kwargs.pop('capabilities', 0)
        det_model_path = kwargs.pop('det_model_path', None)
        rec_model_path = kwargs.pop('rec_model_path', None)
        char_dict_path = kwargs.pop('char_dict_path', None)

        # If det_model_path and rec_model_path are not provided, use model_path (matching Go behavior)
        if not det_model_path:
            det_model_path = model_path
        if not rec_model_path:
            rec_model_path = model_path

        plugin_id = manifest.plugin_id.strip() if manifest.plugin_id else None
        if not plugin_id:
            plugin_id = kwargs.pop('plugin_id', None)
        else:
            # Remove plugin_id from kwargs if manifest has it (to avoid passing it twice)
            kwargs.pop('plugin_id', None)
        if plugin_id:
            plugin_id = plugin_id.strip() if isinstance(plugin_id, str) else plugin_id
            if not plugin_id:
                plugin_id = None

        if not plugin_id:
            plugin_id = cls._get_default_plugin_id()

        device_id = manifest.device_id.strip() if manifest.device_id else None
        if not device_id:
            device_id = kwargs.pop('device_id', None)
        if device_id:
            device_id = device_id.strip() if isinstance(device_id, str) else device_id
            if not device_id:
                device_id = None

        return {
            'model_name': model_name,
            'capabilities': capabilities,
            'det_model_path': det_model_path,
            'rec_model_path': rec_model_path,
            'char_dict_path': char_dict_path,
            'plugin_id': plugin_id,
            'device_id': device_id,
            **kwargs,
        }

    @classmethod
    def _create_instance(cls, **params) -> 'CV':
        """Create CV instance with given parameters."""
        return cls(**params)

    @classmethod
    def _get_default_plugin_id(cls) -> str:
        """Get default plugin ID for CV."""
        return 'metal'

    def infer(self, input_image_path: str) -> 'CVResult':
        """
        Perform CV inference on image.

        Args:
            input_image_path: Path to the input image file.

        Returns:
            CVResult containing detection/classification results.

        Raises:
            NexaError: If inference fails.
        """
        if hasattr(self, '_is_native') and self._is_native:
            # Native implementation
            mlx_result = self._mlx_cv.infer(input_image_path, clear_cache=True)
            return self._convert_mlx_result(mlx_result)

        # C interface implementation
        input_image_path_bytes = input_image_path.encode('utf-8')
        input_image_path_buf = ctypes.create_string_buffer(input_image_path_bytes)

        c_input = ml_CVInferInput(
            input_image_path=ctypes.cast(input_image_path_buf, ctypes.c_char_p),
        )

        c_output = ml_CVInferOutput()
        error_code = _lib.ml_cv_infer(self._handle, ctypes.pointer(c_input), ctypes.pointer(c_output))
        check_error(error_code)

        return self._extract_result(c_output)

    def _convert_mlx_result(self, mlx_result) -> 'CVResult':
        """Convert MLX CV result to CVResult."""
        from ml import CVResults

        results = []
        if isinstance(mlx_result, CVResults) and mlx_result.results:
            for item in mlx_result.results:
                bbox = None
                if item.bbox:
                    bbox = BoundingBox(
                        x=item.bbox.x,
                        y=item.bbox.y,
                        width=item.bbox.width,
                        height=item.bbox.height,
                    )

                results.append(
                    CVResultItem(
                        image_paths=item.image_paths or [],
                        class_id=item.class_id,
                        confidence=item.confidence,
                        bbox=bbox,
                        text=item.text or '',
                        embedding=item.embedding or [],
                        embedding_dim=item.embedding_dim,
                    )
                )

        return CVResult(results=results)

    def _extract_result(self, c_output: ml_CVInferOutput) -> 'CVResult':
        """Extract CVResult from C output structure."""
        if not (c_output.results and c_output.result_count > 0):
            return CVResult(results=[])

        results = []
        for i in range(c_output.result_count):
            result = c_output.results[i]
            image_paths = []
            if result.image_paths and result.image_count > 0:
                image_paths_array = ctypes.cast(
                    result.image_paths, ctypes.POINTER(ctypes.c_char_p * result.image_count)
                ).contents
                for j in range(result.image_count):
                    img_path_ptr = image_paths_array[j]
                    if img_path_ptr:
                        image_paths.append(ctypes.string_at(img_path_ptr).decode('utf-8'))

            text = ''
            if result.text:
                text = ctypes.string_at(result.text).decode('utf-8')

            embedding = []
            if result.embedding and result.embedding_dim > 0:
                embedding_array = ctypes.cast(
                    result.embedding, ctypes.POINTER(ctypes.c_float * result.embedding_dim)
                ).contents
                embedding = [float(embedding_array[j]) for j in range(result.embedding_dim)]

            bbox = BoundingBox(
                x=float(result.bbox.x),
                y=float(result.bbox.y),
                width=float(result.bbox.width),
                height=float(result.bbox.height),
            )

            results.append(
                CVResultItem(
                    image_paths=image_paths,
                    class_id=int(result.class_id),
                    confidence=float(result.confidence),
                    bbox=bbox,
                    text=text,
                    embedding=embedding,
                    embedding_dim=int(result.embedding_dim),
                )
            )

        return CVResult(results=results)


class BoundingBox:
    """Bounding box structure."""

    def __init__(self, x: float, y: float, width: float, height: float):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __repr__(self) -> str:
        return f'BoundingBox(x={self.x}, y={self.y}, width={self.width}, height={self.height})'


class CVResultItem:
    """Single CV result item."""

    def __init__(
        self,
        image_paths: List[str],
        class_id: int,
        confidence: float,
        bbox: Optional[BoundingBox],
        text: str,
        embedding: List[float],
        embedding_dim: int,
    ):
        self.image_paths = image_paths
        self.class_id = class_id
        self.confidence = confidence
        self.bbox = bbox
        self.text = text
        self.embedding = embedding
        self.embedding_dim = embedding_dim

    def __repr__(self) -> str:
        text_preview = self.text[:30] if self.text else None
        return f'CVResultItem(class_id={self.class_id}, confidence={self.confidence:.2f}, text={text_preview})'


class CVResult:
    """Result of CV inference."""

    def __init__(self, results: List[CVResultItem]):
        self.results = results

    def __repr__(self) -> str:
        return f'CVResult(results={len(self.results)})'



"""
LLM functions and structures for nexa-sdk.
"""

import ctypes

from ._lib import _lib
from .error import ml_log_callback

# Callback types
ml_token_callback = ctypes.CFUNCTYPE(ctypes.c_bool, ctypes.c_char_p, ctypes.c_void_p)


# ============================================================================
# C Structure Definitions
# ============================================================================


class ml_SamplerConfig(ctypes.Structure):
    """Text generation sampling parameters."""

    _fields_ = [
        ('temperature', ctypes.c_float),
        ('top_p', ctypes.c_float),
        ('top_k', ctypes.c_int32),
        ('min_p', ctypes.c_float),
        ('repetition_penalty', ctypes.c_float),
        ('presence_penalty', ctypes.c_float),
        ('frequency_penalty', ctypes.c_float),
        ('seed', ctypes.c_int32),
        ('grammar_path', ctypes.c_char_p),
        ('grammar_string', ctypes.c_char_p),
        ('enable_json', ctypes.c_bool),
    ]


class ml_GenerationConfig(ctypes.Structure):
    """LLM / VLM generation configuration."""

    _fields_ = [
        ('max_tokens', ctypes.c_int32),
        ('stop', ctypes.POINTER(ctypes.c_char_p)),
        ('stop_count', ctypes.c_int32),
        ('n_past', ctypes.c_int32),
        ('sampler_config', ctypes.POINTER(ml_SamplerConfig)),
        ('image_paths', ctypes.POINTER(ctypes.c_char_p)),
        ('image_count', ctypes.c_int32),
        ('image_max_length', ctypes.c_int32),
        ('audio_paths', ctypes.POINTER(ctypes.c_char_p)),
        ('audio_count', ctypes.c_int32),
    ]


class ml_ModelConfig(ctypes.Structure):
    """LLM / VLM model configuration."""

    _fields_ = [
        ('n_ctx', ctypes.c_int32),
        ('n_threads', ctypes.c_int32),
        ('n_threads_batch', ctypes.c_int32),
        ('n_batch', ctypes.c_int32),
        ('n_ubatch', ctypes.c_int32),
        ('n_seq_max', ctypes.c_int32),
        ('n_gpu_layers', ctypes.c_int32),
        ('chat_template_path', ctypes.c_char_p),
        ('chat_template_content', ctypes.c_char_p),
        ('system_prompt', ctypes.c_char_p),
        ('enable_sampling', ctypes.c_bool),
        ('grammar_str', ctypes.c_char_p),
        ('max_tokens', ctypes.c_int32),
        ('enable_thinking', ctypes.c_bool),
        ('verbose', ctypes.c_bool),
        ('qnn_model_folder_path', ctypes.c_char_p),
        ('qnn_lib_folder_path', ctypes.c_char_p),
    ]


class ml_LLM(ctypes.Structure):
    """Opaque LLM handle."""

    pass


class ml_LlmCreateInput(ctypes.Structure):
    """Input structure for creating an LLM instance."""

    _fields_ = [
        ('model_name', ctypes.c_char_p),
        ('model_path', ctypes.c_char_p),
        ('tokenizer_path', ctypes.c_char_p),
        ('config', ml_ModelConfig),
        ('plugin_id', ctypes.c_char_p),
        ('device_id', ctypes.c_char_p),
        ('license_id', ctypes.c_char_p),
        ('license_key', ctypes.c_char_p),
    ]


class ml_LlmChatMessage(ctypes.Structure):
    """Chat message structure."""

    _fields_ = [
        ('role', ctypes.c_char_p),
        ('content', ctypes.c_char_p),
    ]


class ml_LlmApplyChatTemplateInput(ctypes.Structure):
    """Input structure for applying chat template."""

    _fields_ = [
        ('messages', ctypes.POINTER(ml_LlmChatMessage)),
        ('message_count', ctypes.c_int32),
        ('tools', ctypes.c_char_p),
        ('enable_thinking', ctypes.c_bool),
        ('add_generation_prompt', ctypes.c_bool),
    ]


class ml_LlmApplyChatTemplateOutput(ctypes.Structure):
    """Output structure for applying chat template."""

    _fields_ = [
        ('formatted_text', ctypes.c_void_p),
    ]


class ml_ProfileData(ctypes.Structure):
    """Profile data structure for performance metrics."""

    _fields_ = [
        ('ttft', ctypes.c_int64),
        ('prompt_time', ctypes.c_int64),
        ('decode_time', ctypes.c_int64),
        ('prompt_tokens', ctypes.c_int64),
        ('generated_tokens', ctypes.c_int64),
        ('audio_duration', ctypes.c_int64),
        ('prefill_speed', ctypes.c_double),
        ('decoding_speed', ctypes.c_double),
        ('real_time_factor', ctypes.c_double),
        ('stop_reason', ctypes.c_char_p),
    ]


class ml_LlmGenerateInput(ctypes.Structure):
    """Input structure for streaming text generation."""

    _fields_ = [
        ('prompt_utf8', ctypes.c_char_p),
        ('config', ctypes.POINTER(ml_GenerationConfig)),
        ('on_token', ml_token_callback),
        ('user_data', ctypes.c_void_p),
    ]


class ml_LlmGenerateOutput(ctypes.Structure):
    """Output structure for streaming text generation."""

    _fields_ = [
        ('full_text', ctypes.c_void_p),
        ('profile_data', ml_ProfileData),
    ]


class ml_KvCacheSaveInput(ctypes.Structure):
    """Input structure for saving KV cache."""

    _fields_ = [
        ('path', ctypes.c_char_p),
    ]


class ml_KvCacheSaveOutput(ctypes.Structure):
    """Output structure for saving KV cache."""

    _fields_ = [
        ('reserved', ctypes.c_void_p),
    ]


class ml_KvCacheLoadInput(ctypes.Structure):
    """Input structure for loading KV cache."""

    _fields_ = [
        ('path', ctypes.c_char_p),
    ]


class ml_KvCacheLoadOutput(ctypes.Structure):
    """Output structure for loading KV cache."""

    _fields_ = [
        ('reserved', ctypes.c_void_p),
    ]


# ============================================================================
# Function Signatures
# ============================================================================

_lib.ml_llm_create.argtypes = [ctypes.POINTER(ml_LlmCreateInput), ctypes.POINTER(ctypes.POINTER(ml_LLM))]
_lib.ml_llm_create.restype = ctypes.c_int32

_lib.ml_llm_destroy.argtypes = [ctypes.POINTER(ml_LLM)]
_lib.ml_llm_destroy.restype = ctypes.c_int32

_lib.ml_llm_reset.argtypes = [ctypes.POINTER(ml_LLM)]
_lib.ml_llm_reset.restype = ctypes.c_int32

_lib.ml_llm_save_kv_cache.argtypes = [
    ctypes.POINTER(ml_LLM),
    ctypes.POINTER(ml_KvCacheSaveInput),
    ctypes.POINTER(ml_KvCacheSaveOutput),
]
_lib.ml_llm_save_kv_cache.restype = ctypes.c_int32

_lib.ml_llm_load_kv_cache.argtypes = [
    ctypes.POINTER(ml_LLM),
    ctypes.POINTER(ml_KvCacheLoadInput),
    ctypes.POINTER(ml_KvCacheLoadOutput),
]
_lib.ml_llm_load_kv_cache.restype = ctypes.c_int32

_lib.ml_llm_apply_chat_template.argtypes = [
    ctypes.POINTER(ml_LLM),
    ctypes.POINTER(ml_LlmApplyChatTemplateInput),
    ctypes.POINTER(ml_LlmApplyChatTemplateOutput),
]
_lib.ml_llm_apply_chat_template.restype = ctypes.c_int32

_lib.ml_llm_generate.argtypes = [
    ctypes.POINTER(ml_LLM),
    ctypes.POINTER(ml_LlmGenerateInput),
    ctypes.POINTER(ml_LlmGenerateOutput),
]
_lib.ml_llm_generate.restype = ctypes.c_int32



"""
NexaAI - Python SDK for Nexa AI ML library.

This package provides a Pythonic interface to the nexa-sdk machine learning library,
including support for LLMs, embeddings, and other ML operations.
"""

import atexit

# IMPORTANT: Preload CUDA libraries BEFORE importing _lib
# This ensures CUDA symbols are available when plugins are loaded
from .nexa_sdk._cuda import _find_cuda

_find_cuda()

from ._version import __version__
from .asr import ASR, TranscribeResult
from .core import _init_logging, get_device_list, get_plugin_list, setup_logging
from .cv import CV, BoundingBox, CVResult, CVResultItem
from .diarize import Diarize, DiarizeResult, SpeechSegment
from .embedding import Embedder, EmbedResult
from .image_gen import ImageGen, ImageGenResult
from .internal.types import DownloadProgressInfo, ModelInfo
from .llm import LLM, GenerateResult
from .models import download_model, list_models, remove_model
from .nexa_sdk._lib import _lib
from .nexa_sdk.error import (
    NexaASRAudioFormatError,
    NexaASRError,
    NexaASRLanguageError,
    NexaASRStreamAlreadyActiveError,
    NexaASRStreamBufferFullError,
    NexaASRStreamCallbackError,
    NexaASRStreamInvalidAudioError,
    NexaASRStreamNotStartedError,
    NexaASRTranscriptionError,
    NexaCommonError,
    NexaCVError,
    NexaCVOCRDetectionError,
    NexaCVOCRFailedError,
    NexaCVOCRRecognitionError,
    NexaDiarizeAudioLoadError,
    NexaDiarizeClusteringError,
    NexaDiarizeEmbeddingError,
    NexaDiarizeError,
    NexaDiarizeSegmentationError,
    NexaEmbeddingDimensionError,
    NexaEmbeddingError,
    NexaEmbeddingGenerationError,
    NexaError,
    NexaFileNotFoundError,
    NexaImageGenDimensionError,
    NexaImageGenError,
    NexaImageGenGenerationError,
    NexaImageGenPromptError,
    NexaInvalidInputError,
    NexaLicenseExpiredError,
    NexaLicenseInvalidError,
    NexaLLMError,
    NexaLLMGenerationError,
    NexaLLMGenerationPromptTooLongError,
    NexaLLMTokenizationContextLengthError,
    NexaLLMTokenizationError,
    NexaMemoryAllocationError,
    NexaModelInvalidError,
    NexaModelLoadError,
    NexaNotInitializedError,
    NexaNotSupportedError,
    NexaRerankError,
    NexaRerankFailedError,
    NexaRerankInputError,
    NexaTTSAudioFormatError,
    NexaTTSError,
    NexaTTSSynthesisError,
    NexaTTSVoiceError,
    NexaUnknownError,
    NexaVLMAudioFormatError,
    NexaVLMAudioLoadError,
    NexaVLMError,
    NexaVLMGenerationError,
    NexaVLMImageFormatError,
    NexaVLMImageLoadError,
    check_error,
    ml_get_error_message,
)
from .nexa_sdk.types import (
    GenerationConfig,
    KvCacheLoadInput,
    KvCacheSaveInput,
    LlmChatMessage,
    ModelConfig,
    ProfileData,
    SamplerConfig,
    VlmChatMessage,
    VlmContent,
)
from .rerank import Reranker, RerankResult
from .tts import TTS, SynthesizeResult
from .vlm import VLM


def version() -> str:
    """Get Python package version."""
    return __version__


def nexa_version() -> str:
    """Get Nexa SDK (C library) version."""
    return _lib.ml_version().decode('utf-8')


_init_logging()

check_error(_lib.ml_init())
atexit.register(lambda: check_error(_lib.ml_deinit()))

__all__ = [
    # Core functions
    'version',
    'nexa_version',
    'get_plugin_list',
    'get_device_list',
    # Logging
    'setup_logging',
    # Model management
    'download_model',
    'list_models',
    'remove_model',
    'ModelInfo',
    'DownloadProgressInfo',
    'FileProgressInfo',
    # LLM
    'LLM',
    'GenerateResult',
    # VLM
    'VLM',
    # ASR
    'ASR',
    'TranscribeResult',
    # TTS
    'TTS',
    'SynthesizeResult',
    # Embedding
    'Embedder',
    'EmbedResult',
    # Rerank
    'Reranker',
    'RerankResult',
    # Diarize
    'Diarize',
    'DiarizeResult',
    'SpeechSegment',
    # CV
    'CV',
    'CVResult',
    'CVResultItem',
    'BoundingBox',
    # ImageGen
    'ImageGen',
    'ImageGenResult',
    # Types
    'SamplerConfig',
    'GenerationConfig',
    'ModelConfig',
    'LlmChatMessage',
    'VlmChatMessage',
    'VlmContent',
    'ProfileData',
    'KvCacheSaveInput',
    'KvCacheLoadInput',
    # Errors
    'NexaError',
    'NexaCommonError',
    'NexaUnknownError',
    'NexaInvalidInputError',
    'NexaMemoryAllocationError',
    'NexaFileNotFoundError',
    'NexaNotInitializedError',
    'NexaNotSupportedError',
    'NexaModelLoadError',
    'NexaModelInvalidError',
    'NexaLicenseInvalidError',
    'NexaLicenseExpiredError',
    'NexaLLMError',
    'NexaLLMTokenizationError',
    'NexaLLMTokenizationContextLengthError',
    'NexaLLMGenerationError',
    'NexaLLMGenerationPromptTooLongError',
    'NexaVLMError',
    'NexaVLMImageLoadError',
    'NexaVLMImageFormatError',
    'NexaVLMAudioLoadError',
    'NexaVLMAudioFormatError',
    'NexaVLMGenerationError',
    'NexaEmbeddingError',
    'NexaEmbeddingGenerationError',
    'NexaEmbeddingDimensionError',
    'NexaRerankError',
    'NexaRerankFailedError',
    'NexaRerankInputError',
    'NexaImageGenError',
    'NexaImageGenGenerationError',
    'NexaImageGenPromptError',
    'NexaImageGenDimensionError',
    'NexaASRError',
    'NexaASRTranscriptionError',
    'NexaASRAudioFormatError',
    'NexaASRLanguageError',
    'NexaASRStreamNotStartedError',
    'NexaASRStreamAlreadyActiveError',
    'NexaASRStreamInvalidAudioError',
    'NexaASRStreamBufferFullError',
    'NexaASRStreamCallbackError',
    'NexaTTSError',
    'NexaTTSSynthesisError',
    'NexaTTSVoiceError',
    'NexaTTSAudioFormatError',
    'NexaCVError',
    'NexaCVOCRDetectionError',
    'NexaCVOCRRecognitionError',
    'NexaCVOCRFailedError',
    'NexaDiarizeError',
    'NexaDiarizeAudioLoadError',
    'NexaDiarizeSegmentationError',
    'NexaDiarizeEmbeddingError',
    'NexaDiarizeClusteringError',
    'check_error',
    'ml_get_error_message',
]



"""
Library loader for nexa-sdk C library.
"""

import ctypes
import logging
import os
import sys
from pathlib import Path

logger = logging.getLogger(__name__)


def _find_library() -> Path:
    lib_name = {
        'win32': 'nexa_bridge.dll',
        'darwin': 'libnexa_bridge.dylib',
        'linux': 'libnexa_bridge.so',
    }.get(sys.platform, None)

    if lib_name is None:
        raise RuntimeError(f'Unsupported platform: {sys.platform}')

    lib_paths = [
        Path(__file__).parent.parent.parent.parent.parent / 'build' / 'out' / lib_name,
        Path(os.getenv('NEXA_PLUGIN_PATH', Path(__file__).parent / 'lib')) / lib_name,
    ]

    for lib_path in lib_paths:
        logger.debug(f'Checking library: {lib_path}')
        if lib_path.exists():
            return lib_path

    raise RuntimeError('Could not find nexa bridge library')


# Load the library
_lib_path = _find_library()
logger.debug(f'Loading library from: {_lib_path}')
_lib = ctypes.CDLL(str(_lib_path))



"""
Model management functions for nexaai library.

This module provides functions for downloading, listing, and managing models
from HuggingFace Hub and the local store.
"""

import logging
import os
from typing import Callable, List, Optional

from .internal.model_hub import ensure_model_downloaded
from .internal.store import Store
from .internal.types import DownloadProgressInfo, ModelInfo


def download_model(
    repo_id: str,
    quant_spec: Optional[str] = None,
    token: Optional[str] = os.getenv('HF_TOKEN', None),
    progress_callback: Optional[Callable[[DownloadProgressInfo], None]] = None,
    local_dir: Optional[str] = None,
    enable_transfer: bool = True,
    show_progress: bool = True,
    custom_endpoint: Optional[str] = None,
    force_download: bool = False,
    **kwargs,
) -> None:
    """
    Download a model from HuggingFace Hub.

    Args:
        repo_id: HuggingFace repository ID (e.g., "microsoft/phi-3-mini-4k-instruct")
        quant_spec: Optional quant name (e.g., "Q4_K_M") to download specific quantization
        token: Optional HuggingFace token for private repositories, default to environment variable HF_TOKEN
        progress_callback: Optional callback function for download progress.
            The callback receives a DownloadProgressInfo dict with the following structure:
            {
                'status': str,  # 'idle', 'downloading', 'completed', 'error'
                'error_message': str,  # Only present if status is 'error'
                'progress': {
                    'total_downloaded': int,  # Bytes downloaded
                    'total_size': int,        # Total bytes to download
                    'percentage': float,      # Progress percentage (0-100)
                    'files_active': int,      # Number of files currently downloading
                    'files_total': int,       # Total number of files
                    'known_total': bool       # Whether total size is known
                },
                'speed': {
                    'bytes_per_second': float,  # Download speed in bytes/sec
                    'formatted': str            # Human readable speed (e.g., "1.2 MB/s")
                },
                'formatting': {
                    'downloaded': str,  # Human readable downloaded size
                    'total_size': str   # Human readable total size
                },
                'timing': {
                    'elapsed_seconds': float,  # Time since download started
                    'eta_seconds': float,      # Estimated time remaining
                    'start_time': float        # Download start timestamp
                }
            }
        local_dir: Optional directory to download the model to. If not provided, uses default cache directory.
        enable_transfer: If True (default), enable hf_transfer for faster downloads.
        show_progress: If True (default), show download progress bar in terminal.
        custom_endpoint: Optional custom HuggingFace Hub endpoint URL.
        force_download: If True, force re-download even if the model already exists locally.
        **kwargs: Extra fields to store in manifest:
            - avatar_url: Optional avatar URL for the model author

    Example:
        >>> def progress(info):
        ...     pct = info['progress']['percentage']
        ...     downloaded = info['formatting']['downloaded']
        ...     speed = info['speed']['formatted']
        ...     print(f"Progress: {pct:.1f}% | {downloaded} | {speed}")
        >>> nexaai.download_model("microsoft/phi-3-mini-4k-instruct", progress_callback=progress)
    """
    logger = logging.getLogger('nexaai')

    # Use unified internal function to ensure consistent behavior
    manifest = ensure_model_downloaded(
        repo_id=repo_id,
        quant_spec=quant_spec,
        token=token,
        progress_callback=progress_callback,
        local_dir=local_dir,
        enable_transfer=enable_transfer,
        show_progress=show_progress,
        custom_endpoint=custom_endpoint,
        force_download=force_download,
        **kwargs,
    )

    if manifest is not None:
        logger.info(f'Model {repo_id} is ready (already existed or downloaded successfully)')


def list_models() -> List[ModelInfo]:
    """
    List all models downloaded in the local store.

    Returns:
        List of ModelInfo objects containing information about each model

    Example:
        >>> models = nexaai.list_models()
        >>> for model in models:
        ...     print(f"{model.repo_id}: {model.model_type} ({model.size_bytes / 1024**3:.2f} GB)")
    """
    store = Store.get()
    manifests = store.list_models()
    model_infos = []
    for manifest in manifests:
        local_path = str(store.model_dir_path() / manifest.name)
        model_infos.append(ModelInfo.from_manifest(manifest, local_path=local_path))
    return model_infos


def remove_model(repo_id: str) -> Optional[ModelInfo]:
    """
    Remove a model from the local store.

    Args:
        repo_id: HuggingFace repository ID (e.g., "microsoft/phi-3-mini-4k-instruct")

    Returns:
        ModelInfo of the removed model, or None if the model was not found

    Example:
        >>> model = nexaai.remove_model("microsoft/phi-3-mini-4k-instruct")
        >>> if model:
        ...     print(f"Removed {model.repo_id} ({model.size_bytes} bytes)")
    """
    store = Store.get()

    # Get the manifest before removing to return ModelInfo
    manifest = store.get_manifest(repo_id)
    if manifest is None:
        return None

    local_path = str(store.model_dir_path() / repo_id)
    model_info = ModelInfo.from_manifest(manifest, local_path=local_path)

    if store.remove_model(repo_id):
        return model_info
    return None



import os
import sys
from pathlib import Path
from typing import Optional

# Plugin IDs that support native Python calls
_NATIVE_PLUGIN_IDS = {'metal', 'mlx'}


def is_native_plugin(plugin_id: Optional[str]) -> bool:
    if not plugin_id:
        return False
    return plugin_id.lower() in _NATIVE_PLUGIN_IDS


def find_mlx_py_lib() -> Optional[Path]:
    import logging

    logger = logging.getLogger(__name__)

    lib_paths = [
        Path(__file__).parent.parent / 'nexa_sdk' / 'lib' / 'metal' / 'py-lib',
        Path(os.getenv('NEXA_PLUGIN_PATH', Path(__file__).parent / 'lib')) / 'metal' / 'py-lib',
    ]

    logger.debug('Searching for py-lib in: ')
    for lib_path in lib_paths:
        logger.debug(f' - {lib_path}')
        if lib_path.exists() and lib_path.is_dir():
            logger.debug(f' - found py-lib at: {lib_path}')
            return lib_path
    return None


def setup_mlx_imports() -> bool:
    import logging

    logger = logging.getLogger(__name__)

    mlx_path = find_mlx_py_lib()
    if not mlx_path or not mlx_path.exists():
        logger.debug(f'MLX py-lib not found at {mlx_path}')
        return False

    mlx_path_str = str(mlx_path)
    logger.debug(f'Found MLX py-lib at: {mlx_path_str}')

    if mlx_path_str not in sys.path:
        sys.path.insert(0, mlx_path_str)
        logger.debug(f'Added {mlx_path_str} to sys.path')

    try:
        import ml  # noqa: F401

        return True
    except ImportError as e:
        logger.debug(f'Failed to import MLX modules: {e}')
        if mlx_path_str in sys.path:
            sys.path.remove(mlx_path_str)
        return False


def get_mlx_llm_class():
    from llm.interface import LLM

    return LLM


def get_mlx_vlm_class():
    from vlm.interface import VLM

    return VLM


def get_mlx_embedder_class():
    from embedding.interface import Embedder

    return Embedder


def get_mlx_asr_class():
    from asr.interface import MlxAsr

    return MlxAsr


def get_mlx_cv_class():
    from cv.interface import CVModel

    return CVModel


def get_mlx_tts_class():
    from tts.interface import MlxTts

    return MlxTts


def get_mlx_reranker_class():
    from rerank.interface import Reranker

    return Reranker


def get_mlx_imagegen_class():
    from image_gen.interface import ImageGen

    return ImageGen


def convert_to_mlx_model_config(config):
    import ml

    return ml.ModelConfig(
        n_ctx=config.n_ctx,
        n_threads=config.n_threads,
        n_threads_batch=config.n_threads_batch,
        n_batch=config.n_batch,
        n_ubatch=config.n_ubatch,
        n_seq_max=config.n_seq_max,
        chat_template_path=config.chat_template_path,
        chat_template_content=config.chat_template_content,
    )


def convert_to_mlx_generation_config(config):
    import ml

    sampler_config = None
    if config.sampler_config:
        sampler_config = ml.SamplerConfig(
            temperature=config.sampler_config.temperature,
            top_p=config.sampler_config.top_p,
            top_k=config.sampler_config.top_k,
            min_p=config.sampler_config.min_p,
            repetition_penalty=config.sampler_config.repetition_penalty,
            presence_penalty=config.sampler_config.presence_penalty,
            frequency_penalty=config.sampler_config.frequency_penalty,
            seed=config.sampler_config.seed,
            grammar_path=config.sampler_config.grammar_path,
            grammar_string=config.sampler_config.grammar_string,
        )

    image_paths = None
    if config.image_paths:
        image_paths = tuple(config.image_paths)

    audio_paths = None
    if config.audio_paths:
        audio_paths = tuple(config.audio_paths)

    return ml.GenerationConfig(
        max_tokens=config.max_tokens,
        stop=tuple(config.stop) if config.stop else tuple(),
        n_past=config.n_past,
        sampler_config=sampler_config,
        image_paths=image_paths,
        audio_paths=audio_paths,
    )


def convert_to_mlx_chat_message(message):
    import ml

    return ml.ChatMessage(
        role=message.role,
        content=message.content,
    )


def convert_to_mlx_asr_config(
    timestamps: Optional[str] = None,
    beam_size: int = 5,
    stream: bool = False,
):
    import ml

    return ml.ASRConfig(
        timestamps=timestamps or 'none',
        beam_size=beam_size,
        stream=stream,
    )


def convert_from_mlx_profile_data(profiling_data):
    from ..nexa_sdk.types import ProfileData

    return ProfileData(
        ttft=profiling_data.ttft if hasattr(profiling_data, 'ttft') else 0,
        prompt_time=profiling_data.prompt_time if hasattr(profiling_data, 'prompt_time') else 0,
        decode_time=profiling_data.decode_time if hasattr(profiling_data, 'decode_time') else 0,
        prompt_tokens=profiling_data.prompt_tokens if hasattr(profiling_data, 'prompt_tokens') else 0,
        generated_tokens=profiling_data.generated_tokens if hasattr(profiling_data, 'generated_tokens') else 0,
        audio_duration=profiling_data.audio_duration if hasattr(profiling_data, 'audio_duration') else 0,
        prefill_speed=profiling_data.prefill_speed if hasattr(profiling_data, 'prefill_speed') else 0.0,
        decoding_speed=profiling_data.decoding_speed if hasattr(profiling_data, 'decoding_speed') else 0.0,
        real_time_factor=profiling_data.real_time_factor if hasattr(profiling_data, 'real_time_factor') else 0.0,
        stop_reason=profiling_data.stop_reason if hasattr(profiling_data, 'stop_reason') else None,
    )



"""
TTS functions and structures for nexa-sdk.
"""

import ctypes

from ._lib import _lib
from .error import ml_log_callback
from .llm import ml_ModelConfig, ml_ProfileData

# ============================================================================
# C Structure Definitions
# ============================================================================


class ml_TTSConfig(ctypes.Structure):
    """TTS synthesis configuration."""

    _fields_ = [
        ('voice', ctypes.c_char_p),
        ('speed', ctypes.c_float),
        ('seed', ctypes.c_int32),
        ('sample_rate', ctypes.c_int32),
    ]


class ml_TTSSamplerConfig(ctypes.Structure):
    """TTS sampling parameters."""

    _fields_ = [
        ('temperature', ctypes.c_float),
        ('noise_scale', ctypes.c_float),
        ('length_scale', ctypes.c_float),
    ]


class ml_TTSResult(ctypes.Structure):
    """TTS synthesis result."""

    _fields_ = [
        ('audio_path', ctypes.c_char_p),
        ('duration_seconds', ctypes.c_float),
        ('sample_rate', ctypes.c_int32),
        ('channels', ctypes.c_int32),
        ('num_samples', ctypes.c_int32),
    ]


class ml_TTS(ctypes.Structure):
    """Opaque TTS handle."""

    pass


class ml_TtsCreateInput(ctypes.Structure):
    """Input structure for creating a TTS instance."""

    _fields_ = [
        ('model_name', ctypes.c_char_p),
        ('model_path', ctypes.c_char_p),
        ('config', ml_ModelConfig),
        ('vocoder_path', ctypes.c_char_p),
        ('plugin_id', ctypes.c_char_p),
        ('device_id', ctypes.c_char_p),
    ]


class ml_TtsSynthesizeInput(ctypes.Structure):
    """Input structure for TTS synthesis."""

    _fields_ = [
        ('text_utf8', ctypes.c_char_p),
        ('config', ctypes.POINTER(ml_TTSConfig)),
        ('output_path', ctypes.c_char_p),
    ]


class ml_TtsSynthesizeOutput(ctypes.Structure):
    """Output structure for TTS synthesis."""

    _fields_ = [
        ('result', ml_TTSResult),
        ('profile_data', ml_ProfileData),
    ]


class ml_TtsListAvailableVoicesInput(ctypes.Structure):
    """Input structure for getting available voices."""

    _fields_ = [
        ('reserved', ctypes.c_void_p),
    ]


class ml_TtsListAvailableVoicesOutput(ctypes.Structure):
    """Output structure for getting available voices."""

    _fields_ = [
        ('voice_ids', ctypes.POINTER(ctypes.c_char_p)),
        ('voice_count', ctypes.c_int32),
    ]


# ============================================================================
# Function Signatures
# ============================================================================

_lib.ml_tts_create.argtypes = [ctypes.POINTER(ml_TtsCreateInput), ctypes.POINTER(ctypes.POINTER(ml_TTS))]
_lib.ml_tts_create.restype = ctypes.c_int32

_lib.ml_tts_destroy.argtypes = [ctypes.POINTER(ml_TTS)]
_lib.ml_tts_destroy.restype = ctypes.c_int32

_lib.ml_tts_synthesize.argtypes = [
    ctypes.POINTER(ml_TTS),
    ctypes.POINTER(ml_TtsSynthesizeInput),
    ctypes.POINTER(ml_TtsSynthesizeOutput),
]
_lib.ml_tts_synthesize.restype = ctypes.c_int32

_lib.ml_tts_list_available_voices.argtypes = [
    ctypes.POINTER(ml_TTS),
    ctypes.POINTER(ml_TtsListAvailableVoicesInput),
    ctypes.POINTER(ml_TtsListAvailableVoicesOutput),
]
_lib.ml_tts_list_available_voices.restype = ctypes.c_int32



"""
ImageGen (Image Generation) wrapper for nexaai.
Provides high-level Python API for image generation operations.
"""

import ctypes
import logging
from typing import List, Optional

from .internal.model_loader import ModelLoaderMixin
from .internal.native_plugin import (
    get_mlx_imagegen_class,
    is_native_plugin,
    setup_mlx_imports,
)
from .nexa_sdk._lib import _lib
from .nexa_sdk.error import check_error
from .nexa_sdk.image_gen import (
    ml_ImageGen,
    ml_ImageGenCreateInput,
    ml_ImageGenerationConfig,
    ml_ImageGenImg2ImgInput,
    ml_ImageGenOutput,
    ml_ImageGenTxt2ImgInput,
    ml_ImageSamplerConfig,
    ml_SchedulerConfig,
)
from .nexa_sdk.types import ModelConfig

logger = logging.getLogger(__name__)


class ImageGen(ModelLoaderMixin):
    """Image generation model wrapper."""

    def __init__(
        self,
        model_path: str,
        model_name: Optional[str] = None,
        config: Optional[ModelConfig] = None,
        scheduler_config_path: Optional[str] = None,
        plugin_id: Optional[str] = None,
        device_id: Optional[str] = None,
    ):
        """
        Create and initialize an ImageGen instance.

        Args:
            model_path: Path to the model file.
            model_name: Name of the model. If None, uses model_path.
            config: Model configuration. If None, uses default config.
            scheduler_config_path: Path to scheduler configuration file.
            plugin_id: Plugin to use for the model. If None, uses default.
            device_id: Device to use for the model. If None, uses default device.
        """
        if config is None:
            config = ModelConfig()

        if model_name is None:
            model_name = model_path

        if not model_name:
            model_name = model_path

        # Check if we should use native Python calls (for MLX plugin)
        self._is_native = False
        is_native = is_native_plugin(plugin_id)
        logger.debug(f'is_native_plugin({plugin_id}) = {is_native}')
        if is_native:
            logger.info(f'Plugin {plugin_id} is a native plugin, attempting native initialization')
            imports_ok = setup_mlx_imports()
            logger.info(f'setup_mlx_imports() = {imports_ok}')
            if not imports_ok:
                raise RuntimeError(
                    f'Failed to setup MLX imports for native plugin {plugin_id}. '
                    f'Cannot fallback to C interface as it would cause GIL conflicts.'
                )
            try:
                logger.info('Initializing MLX native ImageGen')
                self._init_mlx_native(
                    model_path=model_path,
                    model_name=model_name,
                    scheduler_config_path=scheduler_config_path,
                    plugin_id=plugin_id,
                    device_id=device_id,
                )
                logger.info('MLX native ImageGen initialized successfully')
                self._config = config
                return  # Native initialization successful, skip C interface
            except Exception as e:
                raise RuntimeError(
                    f'Failed to initialize MLX native for plugin {plugin_id}: {e}. '
                    f'Cannot fallback to C interface as it would cause GIL conflicts.'
                ) from e
        else:
            logger.debug(f'Plugin {plugin_id} is not a native plugin, using C interface')

        # C interface initialization (default or fallback)
        c_config = config.to_c_struct()

        model_name_bytes = model_name.encode('utf-8')
        model_path_bytes = model_path.encode('utf-8')
        scheduler_config_path_bytes = scheduler_config_path.encode('utf-8') if scheduler_config_path else None
        plugin_id_str = plugin_id if plugin_id else ''
        plugin_id_buf = ctypes.create_string_buffer(plugin_id_str.encode('utf-8'))
        device_id_buf = ctypes.create_string_buffer(device_id.encode('utf-8')) if device_id else None

        model_name_buf = ctypes.create_string_buffer(model_name_bytes)
        model_path_buf = ctypes.create_string_buffer(model_path_bytes)
        scheduler_config_path_buf = (
            ctypes.create_string_buffer(scheduler_config_path_bytes) if scheduler_config_path_bytes else None
        )

        c_input = ml_ImageGenCreateInput(
            model_name=ctypes.cast(model_name_buf, ctypes.c_char_p),
            model_path=ctypes.cast(model_path_buf, ctypes.c_char_p),
            config=c_config,
            scheduler_config_path=ctypes.cast(scheduler_config_path_buf, ctypes.c_char_p)
            if scheduler_config_path_buf
            else None,
            plugin_id=ctypes.cast(plugin_id_buf, ctypes.c_char_p) if plugin_id_buf else None,
            device_id=ctypes.cast(device_id_buf, ctypes.c_char_p) if device_id_buf else None,
        )

        self._string_refs = [
            model_name_buf,
            model_path_buf,
            scheduler_config_path_buf,
            plugin_id_buf,
            device_id_buf,
        ]

        handle_ptr = ctypes.POINTER(ml_ImageGen)()
        error_code = _lib.ml_imagegen_create(ctypes.pointer(c_input), ctypes.byref(handle_ptr))
        check_error(error_code)

        if not handle_ptr:
            raise RuntimeError('Failed to create ImageGen instance')

        self._handle = handle_ptr
        self._config = config

    def _init_mlx_native(
        self,
        model_path: str,
        model_name: str,
        scheduler_config_path: Optional[str],
        plugin_id: Optional[str],
        device_id: Optional[str],
    ):
        """
        Initialize ImageGen using MLX native Python implementation.

        Args:
            model_path: Path to the model file.
            model_name: Name of the model.
            scheduler_config_path: Path to scheduler configuration file.
            plugin_id: Plugin ID.
            device_id: Device ID (e.g., 'cpu', 'gpu').
        """
        MLXImageGen = get_mlx_imagegen_class()

        # Import ImageGenCreateInput from ml
        from pathlib import Path

        from ml import ImageGenCreateInput as MLXImageGenCreateInput
        from ml import ModelConfig as MLXModelConfig

        # Create MLX ImageGenCreateInput
        mlx_input = MLXImageGenCreateInput(
            model_name=model_name,
            model_path=Path(model_path),
            config=MLXModelConfig(),
            scheduler_config_path=Path(scheduler_config_path) if scheduler_config_path else None,
            plugin_id=plugin_id or 'metal',
            device_id=device_id,
        )

        # Create MLX ImageGen instance
        self._mlx_imagegen = MLXImageGen(input=mlx_input)

        self._is_native = True
        self._handle = None  # No C handle for native implementation

    def __del__(self):
        """Destroy ImageGen instance and free associated resources."""
        if hasattr(self, '_is_native') and self._is_native:
            # Native implementation cleanup
            if hasattr(self, '_mlx_imagegen') and self._mlx_imagegen:
                try:
                    self._mlx_imagegen.destroy()
                except Exception as e:
                    logger.warning(f'Error during MLX native ImageGen cleanup: {e}')
                self._mlx_imagegen = None
        elif hasattr(self, '_handle') and self._handle:
            # C interface cleanup
            try:
                _lib.ml_imagegen_destroy(self._handle)
            except Exception as e:
                logger.warning(f'Error during ImageGen cleanup: {e}')
            self._handle = None

    @classmethod
    def _get_model_type(cls) -> str:
        return 'image_gen'

    @classmethod
    def _extract_model_params_from_manifest(
        cls,
        manifest,
        model_path: str,
        repo_id: str,
        store,
        **kwargs,
    ) -> dict:
        """Extract ImageGen-specific parameters from manifest."""
        model_name = manifest.model_name or repo_id
        if not model_name:
            model_name = repo_id

        scheduler_config_path = None
        if hasattr(manifest, 'scheduler_config_file') and manifest.scheduler_config_file:
            scheduler_config_path_obj = store.modelfile_path(repo_id, manifest.scheduler_config_file.name)
            if scheduler_config_path_obj.exists():
                scheduler_config_path = str(scheduler_config_path_obj)

        plugin_id = manifest.plugin_id.strip() if manifest.plugin_id else None
        if not plugin_id:
            plugin_id = kwargs.pop('plugin_id', None)
        else:
            # Remove plugin_id from kwargs if manifest has it (to avoid passing it twice)
            kwargs.pop('plugin_id', None)
        if plugin_id:
            plugin_id = plugin_id.strip() if isinstance(plugin_id, str) else plugin_id
            if not plugin_id:
                plugin_id = None

        if not plugin_id:
            plugin_id = cls._get_default_plugin_id()

        device_id = manifest.device_id.strip() if manifest.device_id else None
        if not device_id:
            device_id = kwargs.pop('device_id', None)
        if device_id:
            device_id = device_id.strip() if isinstance(device_id, str) else device_id
            if not device_id:
                device_id = None

        return {
            'model_path': str(model_path),
            'model_name': model_name,
            'scheduler_config_path': scheduler_config_path,
            'plugin_id': plugin_id,
            'device_id': device_id,
            **kwargs,
        }

    @classmethod
    def _create_instance(cls, **params) -> 'ImageGen':
        """Create ImageGen instance with given parameters."""
        return cls(**params)

    @classmethod
    def _get_default_plugin_id(cls) -> str:
        """Get default plugin ID for ImageGen."""
        return 'nexa_cuda_cpu_gpu'

    def txt2img(
        self,
        prompt: str,
        output_path: str,
        negative_prompts: Optional[List[str]] = None,
        height: int = 512,
        width: int = 512,
        method: str = 'ddpm',
        steps: int = 50,
        guidance_scale: float = 7.5,
        eta: float = 0.0,
        seed: int = -1,
        strength: float = 1.0,
    ) -> 'ImageGenResult':
        """
        Generate image from text prompt.

        Args:
            prompt: Text prompt for image generation.
            output_path: Path to save the generated image.
            negative_prompts: List of negative prompts.
            height: Image height in pixels.
            width: Image width in pixels.
            method: Sampling method (e.g., 'ddpm', 'ddim').
            steps: Number of diffusion steps.
            guidance_scale: Guidance scale for classifier-free guidance.
            eta: Eta parameter for DDIM.
            seed: Random seed. -1 for random.
            strength: Strength parameter.

        Returns:
            ImageGenResult containing output image path.

        Raises:
            NexaError: If generation fails.
        """
        if hasattr(self, '_is_native') and self._is_native:
            # Native implementation
            from pathlib import Path

            from ml import ImageGenerationConfig as MLXImageGenerationConfig
            from ml import ImageGenTxt2ImgInput as MLXImageGenTxt2ImgInput
            from ml import ImageSamplerConfig as MLXImageSamplerConfig

            mlx_sampler_config = MLXImageSamplerConfig(
                steps=steps,
                guidance_scale=guidance_scale,
                seed=seed,
            )

            mlx_generation_config = MLXImageGenerationConfig(
                height=height,
                width=width,
                sampler_config=mlx_sampler_config,
                negative_prompts=negative_prompts or [],
                strength=strength,
            )

            mlx_input = MLXImageGenTxt2ImgInput(
                prompt=prompt,
                config=mlx_generation_config,
                output_path=Path(output_path),
            )

            mlx_result = self._mlx_imagegen.txt2img(mlx_input)

            return ImageGenResult(output_image_path=str(mlx_result.output_image_path))

        # C interface implementation
        sampler_config = ml_ImageSamplerConfig(
            method=method.encode('utf-8'),
            steps=ctypes.c_int32(steps),
            guidance_scale=ctypes.c_float(guidance_scale),
            eta=ctypes.c_float(eta),
            seed=ctypes.c_int32(seed),
        )

        scheduler_config = ml_SchedulerConfig(
            type=None,
            num_train_timesteps=0,
            steps_offset=0,
            beta_start=0.0,
            beta_end=0.0,
            beta_schedule=None,
            prediction_type=None,
            timestep_type=None,
            timestep_spacing=None,
            interpolation_type=None,
            config_path=None,
        )

        prompt_bytes = prompt.encode('utf-8')
        prompt_buf = ctypes.create_string_buffer(prompt_bytes)
        prompts_array = (ctypes.c_char_p * 1)(ctypes.cast(prompt_buf, ctypes.c_char_p))

        negative_prompts_ptr = None
        negative_prompt_count = 0
        if negative_prompts:
            negative_prompt_count = len(negative_prompts)
            negative_prompts_array = (ctypes.c_char_p * negative_prompt_count)()
            negative_prompt_bufs = []
            for i, neg_prompt in enumerate(negative_prompts):
                neg_prompt_bytes = neg_prompt.encode('utf-8')
                neg_prompt_buf = ctypes.create_string_buffer(neg_prompt_bytes)
                negative_prompt_bufs.append(neg_prompt_buf)
                negative_prompts_array[i] = ctypes.cast(neg_prompt_buf, ctypes.c_char_p)
            negative_prompts_ptr = ctypes.cast(negative_prompts_array, ctypes.POINTER(ctypes.c_char_p))
            self._negative_prompt_bufs = negative_prompt_bufs

        generation_config = ml_ImageGenerationConfig(
            prompts=ctypes.cast(prompts_array, ctypes.POINTER(ctypes.c_char_p)),
            prompt_count=ctypes.c_int32(1),
            negative_prompts=negative_prompts_ptr,
            negative_prompt_count=ctypes.c_int32(negative_prompt_count),
            height=ctypes.c_int32(height),
            width=ctypes.c_int32(width),
            sampler_config=sampler_config,
            scheduler_config=scheduler_config,
            strength=ctypes.c_float(strength),
        )

        prompt_utf8_bytes = prompt.encode('utf-8')
        prompt_utf8_buf = ctypes.create_string_buffer(prompt_utf8_bytes)
        output_path_bytes = output_path.encode('utf-8')
        output_path_buf = ctypes.create_string_buffer(output_path_bytes)

        c_input = ml_ImageGenTxt2ImgInput(
            prompt_utf8=ctypes.cast(prompt_utf8_buf, ctypes.c_char_p),
            config=ctypes.pointer(generation_config),
            output_path=ctypes.cast(output_path_buf, ctypes.c_char_p),
        )

        c_output = ml_ImageGenOutput()
        error_code = _lib.ml_imagegen_txt2img(self._handle, ctypes.pointer(c_input), ctypes.pointer(c_output))
        check_error(error_code)

        return self._extract_result(c_output)

    def img2img(
        self,
        init_image_path: str,
        prompt: str,
        output_path: str,
        negative_prompts: Optional[List[str]] = None,
        height: int = 512,
        width: int = 512,
        method: str = 'ddpm',
        steps: int = 50,
        guidance_scale: float = 7.5,
        eta: float = 0.0,
        seed: int = -1,
        strength: float = 0.8,
    ) -> 'ImageGenResult':
        """
        Generate image from existing image and text prompt.

        Args:
            init_image_path: Path to the initial image.
            prompt: Text prompt for image generation.
            output_path: Path to save the generated image.
            negative_prompts: List of negative prompts.
            height: Image height in pixels.
            width: Image width in pixels.
            method: Sampling method (e.g., 'ddpm', 'ddim').
            steps: Number of diffusion steps.
            guidance_scale: Guidance scale for classifier-free guidance.
            eta: Eta parameter for DDIM.
            seed: Random seed. -1 for random.
            strength: Strength parameter (0.0-1.0).

        Returns:
            ImageGenResult containing output image path.

        Raises:
            NexaError: If generation fails.
        """
        if hasattr(self, '_is_native') and self._is_native:
            # Native implementation
            from pathlib import Path

            from ml import ImageGenerationConfig as MLXImageGenerationConfig
            from ml import ImageGenImg2ImgInput as MLXImageGenImg2ImgInput
            from ml import ImageSamplerConfig as MLXImageSamplerConfig
            from PIL import Image

            mlx_sampler_config = MLXImageSamplerConfig(
                steps=steps,
                guidance_scale=guidance_scale,
                seed=seed,
            )

            mlx_generation_config = MLXImageGenerationConfig(
                height=height,
                width=width,
                sampler_config=mlx_sampler_config,
                negative_prompts=negative_prompts or [],
                strength=strength,
            )

            init_image = Image.open(init_image_path)

            mlx_input = MLXImageGenImg2ImgInput(
                init_image_path=Path(init_image_path),
                prompt=prompt,
                config=mlx_generation_config,
                output_path=Path(output_path),
            )

            mlx_result = self._mlx_imagegen.img2img(mlx_input)

            return ImageGenResult(output_image_path=str(mlx_result.output_image_path))

        # C interface implementation
        sampler_config = ml_ImageSamplerConfig(
            method=method.encode('utf-8'),
            steps=ctypes.c_int32(steps),
            guidance_scale=ctypes.c_float(guidance_scale),
            eta=ctypes.c_float(eta),
            seed=ctypes.c_int32(seed),
        )

        scheduler_config = ml_SchedulerConfig(
            type=None,
            num_train_timesteps=0,
            steps_offset=0,
            beta_start=0.0,
            beta_end=0.0,
            beta_schedule=None,
            prediction_type=None,
            timestep_type=None,
            timestep_spacing=None,
            interpolation_type=None,
            config_path=None,
        )

        prompt_bytes = prompt.encode('utf-8')
        prompt_buf = ctypes.create_string_buffer(prompt_bytes)
        prompts_array = (ctypes.c_char_p * 1)(ctypes.cast(prompt_buf, ctypes.c_char_p))

        negative_prompts_ptr = None
        negative_prompt_count = 0
        if negative_prompts:
            negative_prompt_count = len(negative_prompts)
            negative_prompts_array = (ctypes.c_char_p * negative_prompt_count)()
            negative_prompt_bufs = []
            for i, neg_prompt in enumerate(negative_prompts):
                neg_prompt_bytes = neg_prompt.encode('utf-8')
                neg_prompt_buf = ctypes.create_string_buffer(neg_prompt_bytes)
                negative_prompt_bufs.append(neg_prompt_buf)
                negative_prompts_array[i] = ctypes.cast(neg_prompt_buf, ctypes.c_char_p)
            negative_prompts_ptr = ctypes.cast(negative_prompts_array, ctypes.POINTER(ctypes.c_char_p))
            self._negative_prompt_bufs = negative_prompt_bufs

        generation_config = ml_ImageGenerationConfig(
            prompts=ctypes.cast(prompts_array, ctypes.POINTER(ctypes.c_char_p)),
            prompt_count=ctypes.c_int32(1),
            negative_prompts=negative_prompts_ptr,
            negative_prompt_count=ctypes.c_int32(negative_prompt_count),
            height=ctypes.c_int32(height),
            width=ctypes.c_int32(width),
            sampler_config=sampler_config,
            scheduler_config=scheduler_config,
            strength=ctypes.c_float(strength),
        )

        init_image_path_bytes = init_image_path.encode('utf-8')
        init_image_path_buf = ctypes.create_string_buffer(init_image_path_bytes)
        prompt_utf8_bytes = prompt.encode('utf-8')
        prompt_utf8_buf = ctypes.create_string_buffer(prompt_utf8_bytes)
        output_path_bytes = output_path.encode('utf-8')
        output_path_buf = ctypes.create_string_buffer(output_path_bytes)

        c_input = ml_ImageGenImg2ImgInput(
            init_image_path=ctypes.cast(init_image_path_buf, ctypes.c_char_p),
            prompt_utf8=ctypes.cast(prompt_utf8_buf, ctypes.c_char_p),
            config=ctypes.pointer(generation_config),
            output_path=ctypes.cast(output_path_buf, ctypes.c_char_p),
        )

        c_output = ml_ImageGenOutput()
        error_code = _lib.ml_imagegen_img2img(self._handle, ctypes.pointer(c_input), ctypes.pointer(c_output))
        check_error(error_code)

        return self._extract_result(c_output)

    def _extract_result(self, c_output: ml_ImageGenOutput) -> 'ImageGenResult':
        """Extract ImageGenResult from C output structure."""
        output_image_path = ''
        if c_output.output_image_path:
            output_image_path = c_output.output_image_path.decode('utf-8')

        return ImageGenResult(output_image_path=output_image_path)


class ImageGenResult:
    """Result of image generation."""

    def __init__(self, output_image_path: str):
        self.output_image_path = output_image_path

    def __repr__(self) -> str:
        return f'ImageGenResult(output_image_path={self.output_image_path})'



"""
ASR (Automatic Speech Recognition) wrapper for nexaai.
Provides high-level Python API for ASR operations.
"""

import ctypes
import logging
from typing import Callable, List, Optional

from .internal.model_loader import ModelLoaderMixin
from .internal.native_plugin import (
    convert_from_mlx_profile_data,
    convert_to_mlx_asr_config,
    get_mlx_asr_class,
    is_native_plugin,
    setup_mlx_imports,
)
from .nexa_sdk._lib import _lib
from .nexa_sdk.asr import (
    ml_ASR,
    ml_asr_transcription_callback,
    ml_ASRConfig,
    ml_AsrCreateInput,
    ml_AsrListSupportedLanguagesInput,
    ml_AsrListSupportedLanguagesOutput,
    ml_AsrStreamBeginInput,
    ml_AsrStreamBeginOutput,
    ml_ASRStreamConfig,
    ml_AsrStreamPushAudioInput,
    ml_AsrStreamStopInput,
    ml_AsrTranscribeInput,
    ml_AsrTranscribeOutput,
)
from .nexa_sdk.error import check_error
from .nexa_sdk.types import ModelConfig, ProfileData

logger = logging.getLogger(__name__)


class ASR(ModelLoaderMixin):
    """Automatic Speech Recognition wrapper."""

    def __init__(
        self,
        model_path: str,
        tokenizer_path: Optional[str] = None,
        model_name: Optional[str] = None,
        config: Optional[ModelConfig] = None,
        language: Optional[str] = None,
        plugin_id: Optional[str] = None,
        device_id: Optional[str] = None,
        license_id: Optional[str] = None,
        license_key: Optional[str] = None,
    ):
        """
        Create and initialize an ASR instance.

        Args:
            model_path: Path to the model file.
            tokenizer_path: Path to the tokenizer file. If None, uses model_path.
            model_name: Name of the model. If None, uses model_path.
            config: Model configuration. If None, uses default config.
            language: Language code for ASR. If None, uses default.
            plugin_id: Plugin to use for the model. If None, uses default.
            device_id: Device to use for the model. If None, uses default device.
            license_id: License ID for loading NPU models.
            license_key: License key for loading NPU models.
        """
        if config is None:
            config = ModelConfig()

        if tokenizer_path is None:
            tokenizer_path = model_path

        if model_name is None:
            model_name = model_path

        if not model_name:
            model_name = model_path

        # Check if we should use native Python calls (for MLX plugin)
        self._is_native = False
        if is_native_plugin(plugin_id):
            if not setup_mlx_imports():
                raise RuntimeError(
                    f'Failed to setup MLX imports for native plugin {plugin_id}. '
                    f'Cannot fallback to C interface as it would cause GIL conflicts.'
                )
            try:
                self._init_mlx_native(
                    model_path=model_path,
                    tokenizer_path=tokenizer_path,
                    model_name=model_name,
                    config=config,
                    language=language,
                    device_id=device_id,
                )
                return  # Native initialization successful, skip C interface
            except Exception as e:
                raise RuntimeError(
                    f'Failed to initialize MLX native for plugin {plugin_id}: {e}. '
                    f'Cannot fallback to C interface as it would cause GIL conflicts.'
                ) from e

        # C interface initialization (default or fallback)
        c_config = config.to_c_struct()

        model_name_bytes = model_name.encode('utf-8')
        model_path_bytes = model_path.encode('utf-8')
        tokenizer_path_bytes = tokenizer_path.encode('utf-8') if tokenizer_path else b''
        language_bytes = language.encode('utf-8') if language else None
        plugin_id_str = plugin_id if plugin_id else ''
        plugin_id_buf = ctypes.create_string_buffer(plugin_id_str.encode('utf-8'))
        device_id_buf = ctypes.create_string_buffer(device_id.encode('utf-8')) if device_id else None
        license_id_buf = ctypes.create_string_buffer(license_id.encode('utf-8')) if license_id else None
        license_key_buf = ctypes.create_string_buffer(license_key.encode('utf-8')) if license_key else None

        model_name_buf = ctypes.create_string_buffer(model_name_bytes)
        model_path_buf = ctypes.create_string_buffer(model_path_bytes)
        tokenizer_path_buf = ctypes.create_string_buffer(tokenizer_path_bytes) if tokenizer_path_bytes else None
        language_buf = ctypes.create_string_buffer(language_bytes) if language_bytes else None

        c_input = ml_AsrCreateInput(
            model_name=ctypes.cast(model_name_buf, ctypes.c_char_p),
            model_path=ctypes.cast(model_path_buf, ctypes.c_char_p),
            tokenizer_path=ctypes.cast(tokenizer_path_buf, ctypes.c_char_p) if tokenizer_path_buf else None,
            config=c_config,
            language=ctypes.cast(language_buf, ctypes.c_char_p) if language_buf else None,
            plugin_id=ctypes.cast(plugin_id_buf, ctypes.c_char_p) if plugin_id_buf else None,
            device_id=ctypes.cast(device_id_buf, ctypes.c_char_p) if device_id_buf else None,
            license_id=ctypes.cast(license_id_buf, ctypes.c_char_p) if license_id_buf else None,
            license_key=ctypes.cast(license_key_buf, ctypes.c_char_p) if license_key_buf else None,
        )

        self._string_refs = [
            model_name_buf,
            model_path_buf,
            tokenizer_path_buf,
            language_buf,
            plugin_id_buf,
            device_id_buf,
            license_id_buf,
            license_key_buf,
        ]

        handle_ptr = ctypes.POINTER(ml_ASR)()
        error_code = _lib.ml_asr_create(ctypes.pointer(c_input), ctypes.byref(handle_ptr))
        check_error(error_code)

        if not handle_ptr:
            raise RuntimeError('Failed to create ASR instance')

        self._handle = handle_ptr
        self._config = config

    def _init_mlx_native(
        self,
        model_path: str,
        tokenizer_path: str,
        model_name: str,
        config: ModelConfig,
        language: Optional[str],
        device_id: Optional[str],
    ):
        """
        Initialize ASR using MLX native Python implementation.

        Args:
            model_path: Path to the model file
            tokenizer_path: Path to the tokenizer file
            model_name: Name of the model
            config: Model configuration
            language: Language code
            device_id: Device ID (e.g., 'cpu', 'gpu')
        """
        MLXASR = get_mlx_asr_class()

        # Create MLX ASR instance
        self._mlx_asr = MLXASR(
            model_path=model_path,
            tokenizer_path=tokenizer_path if tokenizer_path else None,
            language=language,
        )

        self._is_native = True
        self._config = config
        self._handle = None  # No C handle for native implementation

    def __del__(self):
        """Destroy ASR instance and free associated resources."""
        if hasattr(self, '_is_native') and self._is_native:
            # Native implementation cleanup
            if hasattr(self, '_mlx_asr') and self._mlx_asr:
                try:
                    self._mlx_asr.destroy()
                except Exception as e:
                    logger.warning(f'Error during MLX native ASR cleanup: {e}')
                self._mlx_asr = None
        elif hasattr(self, '_handle') and self._handle:
            # C interface cleanup
            try:
                _lib.ml_asr_destroy(self._handle)
            except Exception as e:
                logger.warning(f'Error during ASR cleanup: {e}')
            self._handle = None

    @classmethod
    def _get_model_type(cls) -> str:
        return 'asr'

    @classmethod
    def _extract_model_params_from_manifest(
        cls,
        manifest,
        model_path: str,
        repo_id: str,
        store,
        **kwargs,
    ) -> dict:
        """Extract ASR-specific parameters from manifest."""
        tokenizer_path = None
        if manifest.tokenizer_file.name:
            tokenizer_path_obj = store.modelfile_path(repo_id, manifest.tokenizer_file.name)
            if tokenizer_path_obj.exists():
                tokenizer_path = str(tokenizer_path_obj)

        model_name = manifest.model_name or repo_id
        if not model_name:
            model_name = repo_id

        plugin_id = manifest.plugin_id.strip() if manifest.plugin_id else None
        if not plugin_id:
            plugin_id = kwargs.pop('plugin_id', None)
        else:
            # Remove plugin_id from kwargs if manifest has it (to avoid passing it twice)
            kwargs.pop('plugin_id', None)
        if plugin_id:
            plugin_id = plugin_id.strip() if isinstance(plugin_id, str) else plugin_id
            if not plugin_id:
                plugin_id = None

        if not plugin_id:
            plugin_id = cls._get_default_plugin_id()

        device_id = manifest.device_id.strip() if manifest.device_id else None
        if not device_id:
            device_id = kwargs.pop('device_id', None)
        if device_id:
            device_id = device_id.strip() if isinstance(device_id, str) else device_id
            if not device_id:
                device_id = None

        return {
            'model_path': str(model_path),
            'tokenizer_path': tokenizer_path,
            'model_name': model_name,
            'plugin_id': plugin_id,
            'device_id': device_id,
            **kwargs,
        }

    @classmethod
    def _create_instance(cls, **params) -> 'ASR':
        """Create ASR instance with given parameters."""
        return cls(**params)

    @classmethod
    def _get_default_plugin_id(cls) -> str:
        """Get default plugin ID for ASR."""
        return 'metal'

    def transcribe(
        self,
        audio_path: str,
        language: Optional[str] = None,
        timestamps: Optional[str] = None,
        beam_size: int = 5,
    ) -> 'TranscribeResult':
        """
        Transcribe audio file to text.

        Args:
            audio_path: Path to the audio file.
            language: Language code. If None, uses default.
            timestamps: Timestamp format. If None, no timestamps.
            beam_size: Beam size for decoding.

        Returns:
            TranscribeResult containing transcript and metadata.

        Raises:
            NexaError: If transcription fails.
        """
        if self._is_native:
            return self._mlx_transcribe(audio_path, language, timestamps, beam_size)

        asr_config = ml_ASRConfig(
            timestamps=timestamps.encode('utf-8') if timestamps else None,
            beam_size=ctypes.c_int32(beam_size),
            stream=ctypes.c_bool(False),
        )

        audio_path_bytes = audio_path.encode('utf-8')
        audio_path_buf = ctypes.create_string_buffer(audio_path_bytes)
        language_bytes = language.encode('utf-8') if language else None
        language_buf = ctypes.create_string_buffer(language_bytes) if language_bytes else None

        c_input = ml_AsrTranscribeInput(
            audio_path=ctypes.cast(audio_path_buf, ctypes.c_char_p),
            language=ctypes.cast(language_buf, ctypes.c_char_p) if language_buf else None,
            config=ctypes.pointer(asr_config),
        )

        c_output = ml_AsrTranscribeOutput()
        error_code = _lib.ml_asr_transcribe(self._handle, ctypes.pointer(c_input), ctypes.pointer(c_output))
        check_error(error_code)

        return self._extract_result(c_output)

    def _mlx_transcribe(
        self,
        audio_path: str,
        language: Optional[str] = None,
        timestamps: Optional[str] = None,
        beam_size: int = 5,
    ) -> 'TranscribeResult':
        """Transcribe using MLX native implementation."""
        # Convert config
        mlx_config = convert_to_mlx_asr_config(
            timestamps=timestamps,
            beam_size=beam_size,
            stream=False,
        )

        # Call MLX transcribe
        result = self._mlx_asr.transcribe(
            audio_path=audio_path,
            language=language,
            config=mlx_config,
        )

        # Extract profiling data if available
        profiling_data = None
        if hasattr(self._mlx_asr, 'get_profiling_data'):
            profiling_data = self._mlx_asr.get_profiling_data()

        profile_data = convert_from_mlx_profile_data(profiling_data) if profiling_data else ProfileData()

        return TranscribeResult(
            transcript=result.transcript,
            confidence_scores=list(result.confidence_scores),
            timestamps=[(ts[0], ts[1]) for ts in result.timestamps],
            profile_data=profile_data,
        )

    def _extract_result(self, c_output: ml_AsrTranscribeOutput) -> 'TranscribeResult':
        """Extract TranscribeResult from C output structure."""
        result = c_output.result
        transcript = ''
        if result.transcript:
            transcript = result.transcript.decode('utf-8')

        confidence_scores = []
        if result.confidence_scores and result.confidence_count > 0:
            confidence_scores = [result.confidence_scores[i] for i in range(result.confidence_count)]

        timestamps = []
        if result.timestamps and result.timestamp_count > 0:
            timestamps = [result.timestamps[i] for i in range(result.timestamp_count)]

        profile_data = ProfileData.from_c_struct(c_output.profile_data)
        return TranscribeResult(
            transcript=transcript,
            confidence_scores=confidence_scores,
            timestamps=timestamps,
            profile_data=profile_data,
        )

    def list_supported_languages(self) -> List[str]:
        """
        List supported languages.

        Returns:
            List of language codes.

        Raises:
            NexaError: If query fails.
        """
        if self._is_native:
            return self._mlx_asr.list_supported_languages()

        c_input = ml_AsrListSupportedLanguagesInput(reserved=None)
        c_output = ml_AsrListSupportedLanguagesOutput()

        error_code = _lib.ml_asr_list_supported_languages(
            self._handle,
            ctypes.pointer(c_input),
            ctypes.pointer(c_output),
        )
        check_error(error_code)

        languages = []
        if c_output.language_codes and c_output.language_count > 0:
            for i in range(c_output.language_count):
                lang_ptr = c_output.language_codes[i]
                if lang_ptr:
                    languages.append(lang_ptr.decode('utf-8'))
            _lib.ml_free(c_output.language_codes)

        return languages

    def stream_begin(
        self,
        language: Optional[str] = None,
        on_transcription: Optional[Callable[[str], None]] = None,
        chunk_duration: float = 0.5,
        overlap_duration: float = 0.1,
        sample_rate: int = 16000,
        max_queue_size: int = 10,
        buffer_size: int = 4096,
        timestamps: Optional[str] = None,
        beam_size: int = 5,
    ) -> None:
        """
        Begin streaming ASR transcription.

        Args:
            language: Language code. If None, uses default.
            on_transcription: Callback function for transcription results.
            chunk_duration: Audio chunk duration in seconds.
            overlap_duration: Overlap duration in seconds.
            sample_rate: Audio sample rate.
            max_queue_size: Maximum queue size.
            buffer_size: Buffer size.
            timestamps: Timestamp format. If None, no timestamps.
            beam_size: Beam size for decoding.

        Raises:
            NexaError: If stream begin fails.
        """
        stream_config = ml_ASRStreamConfig(
            chunk_duration=ctypes.c_float(chunk_duration),
            overlap_duration=ctypes.c_float(overlap_duration),
            sample_rate=ctypes.c_int32(sample_rate),
            max_queue_size=ctypes.c_int32(max_queue_size),
            buffer_size=ctypes.c_int32(buffer_size),
            timestamps=timestamps.encode('utf-8') if timestamps else None,
            beam_size=ctypes.c_int32(beam_size),
        )

        transcription_callback = None
        if on_transcription:

            def _transcription_callback(transcript_ptr: ctypes.c_char_p, user_data_ptr: ctypes.c_void_p) -> None:
                if transcript_ptr:
                    transcript = transcript_ptr.decode('utf-8')
                    on_transcription(transcript)

            transcription_callback = ml_asr_transcription_callback(_transcription_callback)

        language_bytes = language.encode('utf-8') if language else None
        language_buf = ctypes.create_string_buffer(language_bytes) if language_bytes else None

        c_input = ml_AsrStreamBeginInput(
            stream_config=ctypes.pointer(stream_config),
            language=ctypes.cast(language_buf, ctypes.c_char_p) if language_buf else None,
            on_transcription=transcription_callback,
            user_data=None,
        )

        c_output = ml_AsrStreamBeginOutput()
        error_code = _lib.ml_asr_stream_begin(self._handle, ctypes.pointer(c_input), ctypes.pointer(c_output))
        check_error(error_code)

        self._transcription_callback_ref = transcription_callback

    def stream_push_audio(self, audio_data: List[float]) -> None:
        """
        Push audio data for streaming transcription.

        Args:
            audio_data: List of audio samples (float values).

        Raises:
            NexaError: If push fails.
        """
        length = len(audio_data)
        audio_array = (ctypes.c_float * length)(*audio_data)

        c_input = ml_AsrStreamPushAudioInput(
            audio_data=ctypes.cast(audio_array, ctypes.POINTER(ctypes.c_float)),
            length=ctypes.c_int32(length),
        )

        error_code = _lib.ml_asr_stream_push_audio(self._handle, ctypes.pointer(c_input))
        check_error(error_code)

    def stream_stop(self, graceful: bool = True) -> None:
        """
        Stop streaming transcription.

        Args:
            graceful: If True, wait for current processing to complete.

        Raises:
            NexaError: If stop fails.
        """
        c_input = ml_AsrStreamStopInput(graceful=ctypes.c_bool(graceful))
        error_code = _lib.ml_asr_stream_stop(self._handle, ctypes.pointer(c_input))
        check_error(error_code)


class TranscribeResult:
    """Result of ASR transcription."""

    def __init__(
        self,
        transcript: str,
        confidence_scores: List[float],
        timestamps: List[float],
        profile_data: ProfileData,
    ):
        self.transcript = transcript
        self.confidence_scores = confidence_scores
        self.timestamps = timestamps
        self.profile_data = profile_data

    def __str__(self) -> str:
        return self.transcript

    def __repr__(self) -> str:
        transcript_preview = self.transcript[:50] + '...' if len(self.transcript) > 50 else self.transcript
        return (
            f'TranscribeResult(transcript={transcript_preview}, '
            f'confidence_scores={len(self.confidence_scores)}, '
            f'timestamps={len(self.timestamps)})'
        )
